# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

from .mpi import MPITestCase

from ..todmap import OpPointingHpix, TODHpixSpiral, pysm

from ._helpers import create_outdir, create_distdata, uniform_chunks

if pysm is not None:
    from ..todmap import PySMSky, OpSimPySM

from ..dist import distribute_uniform


def setup_toast_observations(self, nside):
    fixture_name = os.path.splitext(os.path.basename(__file__))[0]
    self.outdir = create_outdir(self.comm, fixture_name)

    # Create one observation per group, and each observation will have
    # a fixed number of detectors and one chunk per process.
    self.data = create_distdata(self.comm, obs_per_group=1)

    # Create a single detector.  We could use more, but that would change
    # the results compared to previous versions of these unit tests.
    self.ndet = 1
    self.dets = {"fake_0A": np.array([0.0, 0.0, 0.0, 1.0])}

    # Use one TOD sample per healpix pixel with the "spiral" pointing.
    self.nside = nside
    self.totsamp = 12 * self.nside ** 2
    self.rate = 1.0

    # Chunks - one per process.
    chunks = uniform_chunks(self.totsamp, nchunk=self.data.comm.group_size)

    # Populate the observations (one per group)

    tod = TODHpixSpiral(
        self.data.comm.comm_group,
        self.dets,
        self.totsamp,
        detranks=1,
        firsttime=0.0,
        rate=self.rate,
        nside=self.nside,
        sampsizes=chunks,
    )

    self.data.obs[0]["tod"] = tod

    return


class OpSimPySMTest(MPITestCase):
    def setUp(self):
        setup_toast_observations(self, nside=8)

    def tearDown(self):
        del self.data

    def test_pysm_local_pix(self):
        npix = 12 * self.nside * self.nside

        comm_size = 1 if self.comm is None else self.comm.size
        comm_rank = 0 if self.comm is None else self.comm.rank
        local_start, nlocal = distribute_uniform(npix, comm_size)[comm_rank]
        local_pixels = np.arange(nlocal, dtype=np.int64)
        local_pixels += local_start

        # construct the PySM operator.  Pass in information needed by PySM...

        pysm_sky_config = ["s1", "f1", "a1"]
        bandpasses = {
            "1a": (np.linspace(20, 25, 10), np.ones(10)),
            "1b": (np.linspace(21, 26, 10), np.ones(10)),
            "2a": (np.linspace(18, 23, 10), np.ones(10)),
            "2b": (np.linspace(19, 24, 10), np.ones(10)),
        }
        op = PySMSky(
            pixel_indices=local_pixels,
            nside=self.nside,
            pysm_sky_config=pysm_sky_config,
            units="uK_RJ",
            comm=self.comm,
        )
        local_map = {}  # it should be Cache in production
        op.exec(local_map, out="sky", bandpasses=bandpasses)
        # Script to generate the expected output
        # https://gist.github.com/zonca/56ff738c3d163ee17e3378c34e17a0c3

        if comm_rank == 0:
            np.testing.assert_almost_equal(
                local_map["sky_1a"][0, :3],
                np.array([95.15288056, 76.09502754, 87.41419261]),
                decimal=1,
            )

        if comm_rank == comm_size - 1:
            np.testing.assert_almost_equal(
                local_map["sky_1b"][2, -3:],
                np.array([1.3479588, -0.05170135, -3.72562926]),
                decimal=2,
            )

        return

    def test_pysm_ring_distribution(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank

        # construct the PySM operator.  Pass in information needed by PySM...

        pysm_sky_config = ["s1", "f1", "a1"]
        bandpasses = {
            "1a": (np.linspace(20, 25, 10), np.ones(10)),
            "1b": (np.linspace(21, 26, 10), np.ones(10)),
            "2a": (np.linspace(18, 23, 10), np.ones(10)),
            "2b": (np.linspace(19, 24, 10), np.ones(10)),
        }
        op = PySMSky(
            nside=self.nside,
            pysm_sky_config=pysm_sky_config,
            units="uK_RJ",
            comm=self.comm,
        )
        local_map = {}  # it should be Cache in production
        op.exec(local_map, out="sky", bandpasses=bandpasses)

        if rank == 0:
            # The first process has the first few pixels
            # Expected values computed with PySM 2, integration
            # method is different so they agree only to 1 decimal
            # PySM 3 uses trapz, PySM 2 rectangular but normalizes
            # the weights with trapz
            np.testing.assert_almost_equal(
                local_map["sky_1a"][0, :3],
                np.array([95.15288056, 76.09502754, 87.41419261]),
                decimal=1,
            )

        if rank == 0:
            # The first process has the symmetric rings at the end of the map
            np.testing.assert_almost_equal(
                local_map["sky_1b"][2, -3:],
                np.array([1.3479588, -0.05170135, -3.72562926]),
                decimal=2,
            )

        return

    def test_op_pysm_nosmooth(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # expand the pointing into a low-res pointing matrix
        pointing = OpPointingHpix(nside=self.nside, nest=False, mode="IQU")
        pointing.exec(self.data)

        focalplane = {
            "fake_0A": {
                "bandcenter_ghz": 22.5,
                "bandwidth_ghz": 5,
                "fmin": 1e-05,
                "fwhm": 0.16666666666666666,
            }
        }
        op_sim_pysm = OpSimPySM(
            self.data,
            comm=self.comm,
            out="signal",
            pysm_model=["a1", "f1", "s1"],
            focalplanes=[focalplane],
            apply_beam=False,
            nest=False,
            units="uK_RJ",
        )

        op_sim_pysm.exec(self.data)

        tod = self.data.obs[0]["tod"]
        rescanned_tod = tod.cache.reference("signal_fake_0A")
        pix = tod.cache.reference("pixels_fake_0A")
        weights = tod.cache.reference("weights_fake_0A")

        # compare with maps computes with PySM standalone running
        # the test_mpi.py script from the PySM repository

        I = np.array([95.15288056, 76.09502754, 87.41419261])
        Q = np.array([-7.59180159, 1.30317198, -5.93620791])
        U = np.array([-6.31112965, 7.90560375, -6.61861061])
        expected = []
        for i in range(3):
            expected.append(1.0 * I[i] + weights[i][1] * Q[i] + weights[i][2] * U[i])

        if rank == 0:
            np.testing.assert_array_almost_equal(rescanned_tod[:3], expected, decimal=1)

        return


class OpSimPySMTestSmooth(MPITestCase):
    def setUp(self):
        setup_toast_observations(self, nside=64)

    def test_op_pysm_smooth(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # expand the pointing into a low-res pointing matrix
        pointing = OpPointingHpix(nside=self.nside, nest=False, mode="IQU")
        pointing.exec(self.data)

        focalplane = {
            "fake_0A": {"bandcenter_ghz": 22.5, "bandwidth_ghz": 5, "fwhm": 600}
        }  # fwhm is in arcmin
        op_sim_pysm = OpSimPySM(
            self.data,
            comm=self.comm,
            out="signal",
            pysm_model=["a1", "f1", "s1"],
            focalplanes=[focalplane],
            apply_beam=True,
            nest=False,
            units="uK_RJ",
        )

        op_sim_pysm.exec(self.data)

        tod = self.data.obs[0]["tod"]
        rescanned_tod = tod.cache.reference("signal_fake_0A")
        weights = tod.cache.reference("weights_fake_0A")

        I = np.array([90.15807463, 88.68902424, 89.27413783])
        Q = np.array([-5.08794779, 4.77494585, -4.68392748])
        U = np.array([-8.96212276, 8.80648313, -8.76468294])
        expected = []
        for i in range(3):
            expected.append(1.0 * I[i] + weights[i][1] * Q[i] + weights[i][2] * U[i])

        # With no MPI the smoothing performed by healpy is a bit different, so we reduce
        # the tolerance
        if rank == 0:
            np.testing.assert_array_almost_equal(
                rescanned_tod[:3], expected, decimal=0 if self.comm is None else 1
            )
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os
import shutil

import numpy as np
import numpy.testing as nt

from ..mpi import MPI
from ..tod import AnalyticNoise, OpSimNoise

from ..todmap import (
    TODGround,
    OpPointingHpix,
    OpSimGradient,
    OpSimScan,
    OpSimScanSynchronousSignal,
)

from ..weather import Weather

from ._helpers import (
    create_outdir,
    create_distdata,
    boresight_focalplane,
    create_weather,
)


class OpSimScanSynchronousSignalTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # one detector per process and a single chunk.

        self.data = create_distdata(self.comm, obs_per_group=1)

        # This serial data will exist separately on each process
        if MPI is None:
            self.data_serial = create_distdata(None, obs_per_group=1)
        else:
            self.data_serial = create_distdata(MPI.COMM_SELF, obs_per_group=1)

        self.ndet = self.data.comm.group_size
        self.rate = 20.0

        # Create detectors with white noise
        self.NET = 5.0

        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet, samplerate=self.rate, fknee=0.0, net=self.NET
        )

        dnames_serial, dquat_serial, _, _, _, _, _, _ = boresight_focalplane(
            1, samplerate=self.rate, fknee=0.0, net=self.NET
        )

        # Samples per observation
        self.totsamp = 100000

        # Pixelization
        nside = 256
        self.sim_nside = nside
        self.map_nside = nside

        # Scan properties
        self.site_lon = "-67:47:10"
        self.site_lat = "-22:57:30"
        self.site_alt = 5200.0
        self.coord = "C"
        self.azmin = 45
        self.azmax = 55
        self.el = 60
        self.scanrate = 1.0
        self.scan_accel = 0.1
        self.CES_start = None

        # Populate the single observation per group

        tod = TODGround(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=self.data.comm.group_size,
            firsttime=0.0,
            rate=self.rate,
            site_lon=self.site_lon,
            site_lat=self.site_lat,
            site_alt=self.site_alt,
            azmin=self.azmin,
            azmax=self.azmax,
            el=self.el,
            coord=self.coord,
            scanrate=self.scanrate,
            scan_accel=self.scan_accel,
            CES_start=self.CES_start,
        )

        tod_serial = TODGround(
            self.data_serial.comm.comm_group,
            dquat_serial,
            self.totsamp,
            detranks=self.data_serial.comm.group_size,
            firsttime=0.0,
            rate=self.rate,
            site_lon=self.site_lon,
            site_lat=self.site_lat,
            site_alt=self.site_alt,
            azmin=self.azmin,
            azmax=self.azmax,
            el=self.el,
            coord=self.coord,
            scanrate=self.scanrate,
            scan_accel=self.scan_accel,
            CES_start=self.CES_start,
        )

        self.common_flag_mask = tod.TURNAROUND

        common_flags = tod.read_common_flags()

        # Number of flagged samples in each observation.  Only the first row
        # of the process grid needs to contribute, since all process columns
        # have identical common flags.
        nflagged = 0
        if (tod.grid_comm_col is None) or (tod.grid_comm_col.rank == 0):
            nflagged += np.sum((common_flags & self.common_flag_mask) != 0)

        # Number of flagged samples across all observations
        self.nflagged = None
        if self.comm is None:
            self.nflagged = nflagged
        else:
            self.nflagged = self.data.comm.comm_world.allreduce(nflagged)

        wfile = os.path.join(self.outdir, "weather.fits")
        if self.comm is None or self.comm.rank == 0:
            create_weather(wfile)
        if self.comm is not None:
            self.comm.barrier()

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["id"] = self.data.comm.group
        self.data.obs[0]["telescope_id"] = 1
        self.data.obs[0]["site"] = "blah"
        self.data.obs[0]["site_id"] = 123
        self.data.obs[0]["weather"] = Weather(wfile, site=123)
        self.data.obs[0]["altitude"] = 2000
        self.data.obs[0]["fpradius"] = 1.0

        self.data_serial.obs[0]["tod"] = tod_serial
        self.data_serial.obs[0]["id"] = self.data.comm.group
        self.data_serial.obs[0]["telescope_id"] = 1
        self.data_serial.obs[0]["site"] = "blah"
        self.data_serial.obs[0]["site_id"] = 123
        self.data_serial.obs[0]["weather"] = Weather(wfile, site=123)
        self.data_serial.obs[0]["altitude"] = 2000
        self.data_serial.obs[0]["fpradius"] = 1.0
        return

    def test_sss(self):
        rank = 0
        do_serial = False
        if self.comm is not None:
            rank = self.comm.rank
            do_serial = True

        freq = None
        cachedir = self.outdir

        sss = OpSimScanSynchronousSignal(
            out="sss",
            realization=0,
            component=123456,
            nside=64,
            fwhm=10,
            lmax=256,
            scale=1e-3,
            power=-1,
        )

        # Do the simulation with the default data distribution and communicator

        sss.exec(self.data)

        # Now do an explicit serial calculation on each process for one detector.

        sss.exec(self.data_serial)

        # Check that the two cases agree on the process which has overlap between them
        tod = self.data.obs[0]["tod"]
        oid = self.data.obs[0]["id"]
        tod_serial = self.data_serial.obs[0]["tod"]
        oid_serial = self.data_serial.obs[0]["id"]
        if oid == oid_serial:
            for d in tod.local_dets:
                if d in tod_serial.local_dets:
                    cname = "sss_{}".format(d)
                    ref = tod.cache.reference(cname)
                    ref_serial = tod_serial.cache.reference(cname)
                    nt.assert_almost_equal(ref[:], ref_serial[:])

        return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import dateutil.parser
import os

import healpy as hp
import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

# from ..tod import OpSimAtmosphere, atm_available_utils


def add_todsatellite_args(parser):
    parser.add_argument(
        "--start-time",
        required=False,
        type=float,
        default=0.0,
        help="The overall start time of the simulation",
    )

    parser.add_argument(
        "--spin-period-min",
        required=False,
        type=float,
        default=10.0,
        help="The period (in minutes) of the rotation about the spin axis",
    )
    parser.add_argument(
        "--spin-angle-deg",
        required=False,
        type=float,
        default=30.0,
        help="The opening angle (in degrees) of the boresight from the spin axis",
    )

    parser.add_argument(
        "--prec-period-min",
        required=False,
        type=float,
        default=50.0,
        help="The period (in minutes) of the rotation about the precession axis",
    )
    parser.add_argument(
        "--prec-angle-deg",
        required=False,
        type=float,
        default=65.0,
        help="The opening angle (in degrees) of the spin axis "
        "from the precession axis",
    )

    parser.add_argument(
        "--obs-time-h",
        required=False,
        type=float,
        default=1.0,
        help="Number of hours in one science observation",
    )
    parser.add_argument(
        "--gap-h",
        required=False,
        type=float,
        default=0.0,
        help="Cooler cycle time in hours between science obs",
    )
    parser.add_argument(
        "--obs-num",
        required=False,
        type=int,
        default=1,
        help="Number of complete observations",
    )

    # The HWP arguments may also be added by other TOD classes
    try:
        parser.add_argument(
            "--hwp-rpm",
            required=False,
            type=np.float,
            help="The rate (in RPM) of the HWP rotation",
        )
    except argparse.ArgumentError:
        pass
    try:
        parser.add_argument(
            "--hwp-step-deg",
            required=False,
            type=np.float,
            help="For stepped HWP, the angle in degrees of each step",
        )
    except argparse.ArgumentError:
        pass
    try:
        parser.add_argument(
            "--hwp-step-time-s",
            required=False,
            type=np.float,
            help="For stepped HWP, the time in seconds between steps",
        )
    except argparse.ArgumentError:
        pass

    return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import astropy.io.fits as pf
import datetime
import numpy as np

from . import rng as rng

from .timing import function_timer


class Weather(object):
    """TOAST Weather objects allow sampling weather parameters.

    The weather parameter distributions are read from site-specific
    TOAST weather files.  The files contain parameter distributions
    for every UTC hour of the day, averaged over months.

    """

    def __init__(self, fname, site=0, realization=0, time=None):
        """Initialize a weather object

        Args:
            fname(str) : FITS file containing the parameter
                distributions.
            site(int) : Site index for the random number generator.
            realization(int) : Initial realization index, may be
                changed later.

        """
        self._fname = fname
        self.site = site
        self.realization = realization
        if time is None:
            self._time = None
            self._year = None
            self._month = None
            self._hour = None
        else:
            self.set_time(time)
        self._varindex = {}

        if self._fname is None:
            # We will simulate fake weather information.
            prob_start = 0.0
            prob_stop = 1.0
            nstep = 101
            self._prob = np.linspace(prob_start, prob_stop, nstep)

            self._monthly_cdf = []
            ivar = 0
            for month in range(12):
                self._monthly_cdf.append([])
                # One entry for every hour
                for hour in range(24):
                    self._monthly_cdf[month].append({})
                # and one entry for every weather variable:
                #  TQI   : ice water
                #  TQL   : liquid water
                #  TQV   : water vapor
                #  QV10M : specific humidity
                #  PS    : surface pressure
                #  TS    : surface temperature
                #  T10M  : air temperature at 10m
                #  U10M  : eastward wind at 10m
                #  V10M  : northward wind at 10m
                name = "TQI"
                self._varindex[name] = 0
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = 0.0
                name = "TQL"
                self._varindex[name] = 1
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = 0.0
                name = "TQV"
                self._varindex[name] = 2
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = 0.2
                name = "QV10M"
                self._varindex[name] = 3
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = 0.00015
                name = "PS"
                self._varindex[name] = 4
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = 58200.0
                name = "TS"
                self._varindex[name] = 5
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = 263.0
                name = "T10M"
                self._varindex[name] = 6
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = 263.0
                name = "U10M"
                self._varindex[name] = 7
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = -3.0
                name = "V10M"
                self._varindex[name] = 8
                for hour in range(24):
                    self._monthly_cdf[month][hour][name] = -7.0
        else:
            hdulist = pf.open(self._fname, "readonly")

            # Build the probability axis of the cumulative distribution
            # function.  The CDF:s stored for every month, hour and variable
            # all assume the same probability axis.
            prob_start = hdulist[1].header["probstrt"]
            prob_stop = hdulist[1].header["probstop"]
            nstep = hdulist[1].header["nstep"]
            self._prob = np.linspace(prob_start, prob_stop, nstep)

            # Load the CDF:s.  One entry per month.
            self._monthly_cdf = []
            ivar = 0
            for month in range(12):
                self._monthly_cdf.append([])
                hdu = hdulist[1 + month]
                # One entry for every hour
                for hour in range(24):
                    self._monthly_cdf[month].append({})
                # and one entry for every weather variable:
                #  TQI   : ice water
                #  TQL   : liquid water
                #  TQV   : water vapor
                #  QV10M : specific humidity
                #  PS    : surface pressure
                #  TS    : surface temperature
                #  T10M  : air temperature at 10m
                #  U10M  : eastward wind at 10m
                #  V10M  : northward wind at 10m
                for col in hdu.columns:
                    name = col.name
                    if name not in self._varindex:
                        self._varindex[name] = ivar
                        ivar += 1
                    for hour in range(24):
                        self._monthly_cdf[month][hour][name] = hdu.data.field(name)[
                            hour
                        ]

            hdulist.close()

        self._reset_vars()

        return

    def set(self, site, realization, time=None):
        """Set the weather object state.

        Args:
            site(int) : Site index.
            realization(int) : Realization index.
            time : POSIX timestamp.

        """
        self.site = site
        self.realization = realization
        if time is not None:
            self.set_time(time)
        else:
            self._reset_vars()
        return

    def _reset_vars(self):
        """Reset the cached random variables."""
        self._ice_water = None
        self._liquid_water = None
        self._pwv = None
        self._humidity = None
        self._surface_pressure = None
        self._surface_temperature = None
        self._air_temperature = None
        self._west_wind = None
        self._south_wind = None

    def set_time(self, time):
        """Set the observing time.

        Args:
            time : POSIX timestamp.

        """
        self._time = time
        self._date = datetime.datetime.utcfromtimestamp(self._time)
        self._doy = self._date.timetuple().tm_yday
        self._year = self._date.year
        self._hour = self._date.hour
        # This is the definition of month used in the weather files
        self._month = int((self._doy - 1) // 30.5)
        self._reset_vars()
        return

    @function_timer
    def _draw(self, name):
        """Return a random parameter value.

        Return a random value for preset variable and time.

        Args:
            name(str): MERRA-2 name for the variable.

        """
        if self._year is None:
            raise RuntimeError(
                "Weather object must be initialized by calling " "set_time(time)"
            )
        # Set the RNG counters for this variable and time
        counter1 = self._varindex[name]
        counter2 = (self._year * 366 + self._doy) * 24 + self._hour
        # Get a uniform random number for inverse sampling
        x = rng.random(
            1,
            sampler="uniform_01",
            key=(self.site, self.realization),
            counter=(counter1, counter2),
        )[0]
        # Sample the variable from the inverse cumulative distribution function
        cdf = self._monthly_cdf[self._month][self._hour][name]

        return np.interp(x, self._prob, cdf)

    @property
    def ice_water(self):
        """Total precipitable ice water [kg/m^2] (also [mm]).

        Ice water column at the observing site at the preset time and
        for the preset realization.

        """
        if self._ice_water is None:
            self._ice_water = self._draw("TQI")
        return self._ice_water

    @property
    def liquid_water(self):
        """Total precipitable liquid water [kg/m^2] (also [mm]).

        Liquid water column at the observing site at the preset time and
        for the preset realization.

        """
        if self._liquid_water is None:
            self._liquid_water = self._draw("TQL")
        return self._liquid_water

    @property
    def pwv(self):
        """Total precipitable water vapor [kg/m^2] (also [mm]).

        Water vapor column at the observing site at the preset time and
        for the preset realization.

        """
        if self._pwv is None:
            self._pwv = self._draw("TQV")
        return self._pwv

    @property
    def humidity(self):
        """10-meter specific humidity [kg/kg]

        Water vapor concentration at the observing site 10 meters above
        ground at the preset time and for the preset realization.

        """
        if self._humidity is None:
            self._humidity = self._draw("QV10M")
        return self._humidity

    @property
    def surface_pressure(self):
        """Surface pressure [Pa].

        Surface at the observing site at the preset time and for the
        preset realization.

        """
        if self._surface_pressure is None:
            self._surface_pressure = self._draw("PS")
        return self._surface_pressure

    @property
    def surface_temperature(self):
        """Surface skin temperature [K].

        Surface temperature at the observing site at the preset time and
        for the preset realization.

        """
        if self._surface_temperature is None:
            self._surface_temperature = self._draw("TS")
        return self._surface_temperature

    @property
    def air_temperature(self):
        """10-meter air temperature [K].

        Air temperature at the observing site 10 meters above ground
        at the preset time and for the preset realization.

        """
        if self._air_temperature is None:
            self._air_temperature = self._draw("T10M")
        return self._air_temperature

    @property
    def west_wind(self):
        """10-meter eastward wind [m/s].

        Eastward wind at the observing site 10 meters above ground
        at the preset time and for the preset realization.

        """
        if self._west_wind is None:
            self._west_wind = self._draw("U10M")
        return self._west_wind

    @property
    def south_wind(self):
        """10-meter northward wind [m/s].

        Northward wind at the observing site 10 meters above ground
        at the preset time and for the preset realization.

        """
        if self._south_wind is None:
            self._south_wind = self._draw("V10M")
        return self._south_wind

    def __repr__(self):
        value = (
            "(Weather : '{}', site = {}, time = {}, year = {}, month = {}, "
            "hour = {}, realization = {})"
            "".format(
                self._fname,
                self.site,
                self._time,
                self._year,
                self._month,
                self._hour,
                self.realization,
            )
        )
        return value



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import copy
import os
import re

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..todmap import OpMadam


def add_madam_args(parser):
    """Add libmadam arguments"""

    parser.add_argument(
        "--madam-prefix",
        required=False,
        default="toast",
        help="Output map prefix",
        dest="mapmaker_prefix",
    )
    parser.add_argument(
        "--madam-iter-max",
        required=False,
        default=1000,
        type=np.int,
        help="Maximum number of CG iterations in Madam",
        dest="mapmaker_iter_max",
    )
    parser.add_argument(
        "--madam-precond-width",
        required=False,
        default=100,
        type=np.int,
        help="Width of the Madam band preconditioner",
        dest="mapmaker_precond_width",
    )
    parser.add_argument(
        "--madam-precond-width-min",
        required=False,
        type=np.int,
        help="Minimum width of the Madam band preconditioner",
    )
    parser.add_argument(
        "--madam-precond-width-max",
        required=False,
        type=np.int,
        help="Maximum width of the Madam band preconditioner",
    )
    parser.add_argument(
        "--madam-baseline-length",
        required=False,
        default=10000.0,
        type=np.float,
        help="Destriping baseline length (seconds)",
        dest="mapmaker_baseline_length",
    )
    parser.add_argument(
        "--madam-baseline-order",
        required=False,
        default=0,
        type=np.int,
        help="Destriping baseline polynomial order",
    )
    parser.add_argument(
        "--madam-noisefilter",
        required=False,
        default=False,
        action="store_true",
        help="Destripe with the noise filter enabled",
        dest="mapmaker_noisefilter",
    )
    parser.add_argument(
        "--madam-parfile", required=False, default=None, help="Madam parameter file"
    )
    parser.add_argument(
        "--madam-mask",
        required=False,
        help="Destriping mask",
        dest="mapmaker_mask",
    )

    parser.add_argument(
        "--madam-allreduce",
        required=False,
        action="store_true",
        help="Use the allreduce communication pattern in Madam",
        dest="madam_allreduce",
    )
    parser.add_argument(
        "--no-madam-allreduce",
        required=False,
        action="store_false",
        help="Do not use the allreduce communication pattern in Madam",
        dest="madam_allreduce",
    )
    parser.set_defaults(madam_allreduce=False)

    parser.add_argument(
        "--madam-concatenate-messages",
        required=False,
        action="store_true",
        help="Use the alltoallv commucation pattern in Madam",
        dest="madam_concatenate_messages",
    )
    parser.add_argument(
        "--no-madam-concatenate-messages",
        required=False,
        action="store_false",
        help="Use the point-to-point communication pattern in Madam",
        dest="madam_concatenate_messages",
    )
    parser.set_defaults(madam_concatenate_messages=True)

    try:
        parser.add_argument(
            "--destripe",
            required=False,
            action="store_true",
            help="Write destriped maps [default]",
            dest="destripe",
        )
        parser.add_argument(
            "--no-destripe",
            required=False,
            action="store_false",
            help="Do not write destriped maps",
            dest="destripe",
        )
        parser.set_defaults(destripe=True)
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--binmap",
            required=False,
            action="store_true",
            help="Write binned maps [default]",
            dest="write_binmap",
        )
        parser.add_argument(
            "--no-binmap",
            required=False,
            action="store_false",
            help="Do not write binned maps",
            dest="write_binmap",
        )
        parser.set_defaults(write_binmap=True)
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--hits",
            required=False,
            action="store_true",
            help="Write hit maps [default]",
            dest="write_hits",
        )
        parser.add_argument(
            "--no-hits",
            required=False,
            action="store_false",
            help="Do not write hit maps",
            dest="write_hits",
        )
        parser.set_defaults(write_hits=True)
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--wcov",
            required=False,
            action="store_true",
            help="Write white noise covariance [default]",
            dest="write_wcov",
        )
        parser.add_argument(
            "--no-wcov",
            required=False,
            action="store_false",
            help="Do not write white noise covariance",
            dest="write_wcov",
        )
        parser.set_defaults(write_wcov=True)
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--wcov-inv",
            required=False,
            action="store_true",
            help="Write inverse white noise covariance [default]",
            dest="write_wcov_inv",
        )
        parser.add_argument(
            "--no-wcov-inv",
            required=False,
            action="store_false",
            help="Do not write inverse white noise covariance",
            dest="write_wcov_inv",
        )
        parser.set_defaults(write_wcov_inv=True)
    except argparse.ArgumentError:
        pass

    parser.add_argument(
        "--conserve-memory",
        dest="conserve_memory",
        required=False,
        action="store_true",
        help="Conserve memory when staging libMadam buffers [default]",
    )
    parser.add_argument(
        "--no-conserve-memory",
        dest="conserve_memory",
        required=False,
        action="store_false",
        help="Do not conserve memory when staging libMadam buffers",
    )
    parser.set_defaults(conserve_memory=True)

    # `nside` may already be added
    try:
        parser.add_argument(
            "--nside", required=False, default=512, type=np.int, help="Healpix NSIDE"
        )
    except argparse.ArgumentError:
        pass
    # `nside_submap` may already be added
    try:
        parser.add_argument(
            "--nside_submap",
            required=False,
            default=16,
            type=np.int,
            help="Number of submaps is 12 * nside_submap ** 2",
        )
    except argparse.ArgumentError:
        pass
    # Common flag mask may already be added
    try:
        parser.add_argument(
            "--common-flag-mask",
            required=False,
            default=1,
            type=np.uint8,
            help="Common flag mask",
        )
    except argparse.ArgumentError:
        pass
    # `sample-rate` may be already added
    try:
        parser.add_argument(
            "--sample-rate",
            required=False,
            default=100.0,
            type=np.float,
            help="Detector sample rate (Hz)",
        )
    except argparse.ArgumentError:
        pass
    return


@function_timer
def setup_madam(args):
    """Create a Madam parameter dictionary.

    Initialize the Madam parameters from the command line arguments.
    """
    pars = {}

    cross = args.nside // 2
    submap = args.nside_submap
    if submap > args.nside:
        submap = args.nside

    pars["temperature_only"] = False
    pars["force_pol"] = True
    pars["kfirst"] = args.destripe
    pars["write_map"] = args.destripe
    pars["write_binmap"] = args.write_binmap
    pars["write_matrix"] = args.write_wcov_inv
    pars["write_wcov"] = args.write_wcov
    pars["write_hits"] = args.write_hits
    pars["nside_cross"] = cross
    pars["nside_submap"] = submap
    if args.madam_concatenate_messages:
        # Collective communication is fast but requires memory
        pars["concatenate_messages"] = True
        if args.madam_allreduce:
            # Every process will allocate a copy of every observed submap.
            pars["allreduce"] = True
        else:
            # Every process will allocate complete send and receive buffers
            pars["allreduce"] = False
    else:
        # Slow but memory-efficient point-to-point communication.  Allocate
        # only enough memory to communicate with one process at a time.
        pars["concatenate_messages"] = False
        pars["allreduce"] = False
    if args.mapmaker_mask:
        pars["file_inmask"] = args.mapmaker_mask
    pars["reassign_submaps"] = True
    pars["pixlim_cross"] = 1e-3
    pars["pixmode_cross"] = 2
    pars["pixlim_map"] = 1e-2
    pars["pixmode_map"] = 2
    # Instead of fixed detector weights, we'll want to use scaled noise
    # PSD:s that include the atmospheric noise
    pars["radiometers"] = True
    pars["noise_weights_from_psd"] = True

    if args.madam_parfile is not None:
        # Parse all available parameters from the supplied
        # Madam parameter file
        pat = re.compile(r"\s*(\S+)\s*=\s*(\S+(\s+\S+)*)\s*")
        comment = re.compile(r"^#.*")
        with open(args.madam_parfile, "r") as f:
            for line in f:
                if comment.match(line) is None:
                    result = pat.match(line)
                    if result is not None:
                        key, value = result.group(1), result.group(2)
                        pars[key] = value

    pars["base_first"] = args.mapmaker_baseline_length
    pars["basis_order"] = args.madam_baseline_order
    # Adaptive preconditioner width
    width_min = args.madam_precond_width_min
    width_max = args.madam_precond_width_max
    if width_min is None:
        # madam-precond-width has a default value
        width_min = args.mapmaker_precond_width
    if width_max is None:
        # madam-precond-width has a default value
        width_max = args.mapmaker_precond_width
    if width_min > width_max:
        # it is not an error for these two to match
        width_min = width_max
    pars["precond_width_min"] = width_min
    pars["precond_width_max"] = width_max
    #
    pars["nside_map"] = args.nside
    if args.mapmaker_noisefilter:
        if args.madam_baseline_order != 0:
            raise RuntimeError(
                "Madam cannot build a noise filter when baseline"
                "order is higher than zero."
            )
        pars["kfilter"] = True
    else:
        pars["kfilter"] = False
    pars["fsample"] = args.sample_rate
    pars["iter_max"] = args.mapmaker_iter_max
    pars["file_root"] = args.mapmaker_prefix

    # Translate boolean values.  Madam knows how to do this but it
    # simplifies pipeline_tools/madam.py

    for key, value in pars.items():
        if value == "T":
            pars[key] = True
        elif value == "F":
            pars[key] = False

    return pars


@function_timer
def apply_madam(
    args,
    comm,
    data,
    madampars,
    outpath,
    detweights,
    cache_name,
    freq=None,
    time_comms=None,
    telescope_data=None,
    first_call=True,
    extra_prefix=None,
    verbose=True,
    bin_only=False,
):
    """Use libmadam to bin and optionally destripe data.

    Bin and optionally destripe all conceivable subsets of the data.

    Args:
        freq (str) :  Frequency identifier to append to the file prefix
        time_comms (iterable) :  Series of disjoint communicators that
            map, e.g., seasons and days.  Each entry is a tuple of
            the form (`name`, `communicator`)
        telescope_data (iterable) : series of disjoint TOAST data
            objects.  Each entry is tuple of the form (`name`, `data`).
        bin_only (bool) :  Disable destriping and only bin the signal,
            Useful when running Madam as a part of a filter+bin pipeline.
    """
    if comm.comm_world is None:
        raise RuntimeError("Madam requires MPI")

    log = Logger.get()
    total_timer = Timer()
    total_timer.start()
    if comm.world_rank == 0 and verbose:
        log.info("Making maps")

    pars = copy.deepcopy(madampars)
    pars["path_output"] = outpath
    if comm.world_rank == 0:
        os.makedirs(outpath, exist_ok=True)
    file_root = pars["file_root"]
    if extra_prefix is not None:
        if len(file_root) > 0 and not file_root.endswith("_"):
            file_root += "_"
        file_root += "{}".format(extra_prefix)
    if freq is not None:
        if len(file_root) > 0 and not file_root.endswith("_"):
            file_root += "_"
        file_root += "{:03}".format(int(freq))

    if not first_call:
        # Only the first MC iteration should produce the hits and
        # white noise matrices
        pars["write_matrix"] = False
        pars["write_wcov"] = False
        pars["write_hits"] = False

    if bin_only:
        pars["kfirst"] = False
        pars["write_map"] = False
        pars["write_binmap"] = True

    # Sanity check, is any of the Madam outputs required?

    outputs = [
        pars["write_map"],
        pars["write_binmap"],
        pars["write_hits"],
        pars["write_wcov"],
        pars["write_matrix"],
    ]
    if not np.any(outputs):
        if comm.world_rank == 0:
            log.info("No Madam outputs requested.  Skipping.")
        return

    if args.mapmaker_noisefilter or not pars["kfirst"]:
        # With the noise filter enabled, we want to enforce continuity
        # across the Observation.  Otherwise we fit each interval
        # separately.
        madam_intervals = None
    else:
        madam_intervals = "intervals"

    madam = OpMadam(
        params=pars,
        detweights=detweights,
        name=cache_name,
        common_flag_mask=args.common_flag_mask,
        purge_tod=False,
        intervals=madam_intervals,
        conserve_memory=args.conserve_memory,
    )

    if "info" in madam.params:
        info = madam.params["info"]
    else:
        info = 3

    if time_comms is None:
        time_comms = [("all", comm.comm_world)]

    if telescope_data is None:
        telescope_data = [("all", data)]

    timer = Timer()
    for time_name, time_comm in time_comms:
        for tele_name, tele_data in telescope_data:
            if len(time_name.split("-")) == 3:
                # Special rules for daily maps
                if not args.do_daymaps:
                    continue
                if len(telescope_data) > 1 and tele_name == "all":
                    # Skip daily maps over multiple telescopes
                    continue
                if first_call:
                    # Do not destripe daily maps
                    kfirst_save = pars["kfirst"]
                    write_map_save = pars["write_map"]
                    write_binmap_save = pars["write_binmap"]
                    pars["kfirst"] = False
                    pars["write_map"] = False
                    pars["write_binmap"] = True

            timer.start()
            madam.params["file_root"] = "{}_telescope_{}_time_{}".format(
                file_root, tele_name, time_name
            )
            if time_comm == comm.comm_world:
                madam.params["info"] = info
            else:
                # Cannot have verbose output from concurrent mapmaking
                madam.params["info"] = 0
            if (time_comm is None or time_comm.rank == 0) and verbose:
                log.info("Mapping {}".format(madam.params["file_root"]))
            madam.exec(tele_data, time_comm)

            if time_comm is not None:
                time_comm.barrier()
            if comm.world_rank == 0 and verbose:
                timer.report_clear("Mapping {}".format(madam.params["file_root"]))

            if len(time_name.split("-")) == 3 and first_call:
                # Restore destriping parameters
                pars["kfirst"] = kfirst_save
                pars["write_map"] = write_map_save
                pars["write_binmap"] = write_binmap_save

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    total_timer.stop()
    if comm.world_rank == 0 and verbose:
        total_timer.report("Madam total")

    return



#!/usr/bin/env python3
"""
Use PySM to pre-compute a detector signal map that we can then scan from
in parallel.
"""

import argparse

import numpy as np

import healpy as hp

import pysm
import pysm.units as u

parser = argparse.ArgumentParser(description="Run PySM to build a sky map.")

parser.add_argument(
    "--output", required=False, default=None, help="The output map FITS file.",
)

parser.add_argument(
    "--nside", required=True, default=None, type=int, help="The map NSIDE resolution",
)

parser.add_argument(
    "--beam_arcmin",
    required=True,
    default=None,
    type=float,
    help="The beam FWHM in arcmin",
)

parser.add_argument(
    "--bandcenter_ghz",
    required=True,
    default=None,
    type=float,
    help="The band center in GHz",
)

parser.add_argument(
    "--bandwidth_ghz",
    required=True,
    default=None,
    type=float,
    help="The band width in GHz",
)

parser.add_argument(
    "--coord",
    required=True,
    default=None,
    type=str,
    help="The coordinate system (G, E, C)",
)

args = parser.parse_args()

outfile = args.output
if outfile is None:
    outfile = "input_sky_band-{:.1f}-{:.1f}_beam-{:.1f}_n{:02d}-{}.fits".format(
        args.bandcenter_ghz,
        args.bandwidth_ghz,
        args.beam_arcmin,
        args.nside,
        args.coord,
    )

print("Working on {}".format(outfile))

print("  Construct Sky...", flush=True)
sky = pysm.Sky(nside=args.nside, preset_strings=["d1", "s1", "f1", "a1"])

freqs = (
    np.array(
        [
            args.bandcenter_ghz - 0.5 * args.bandwidth_ghz,
            args.bandcenter_ghz,
            args.bandcenter_ghz + 0.5 * args.bandwidth_ghz,
        ]
    )
    * u.GHz
)

print("  Get Emission...", flush=True)
map_data = sky.get_emission(freqs)

if args.coord == "G":
    print("  Smoothing...", flush=True)
    smoothed = pysm.apply_smoothing_and_coord_transform(
        map_data, fwhm=args.beam_arcmin * u.arcmin
    )
else:
    to_coord = "G{}".format(args.coord)
    print("  Smoothing and rotating...", flush=True)
    smoothed = pysm.apply_smoothing_and_coord_transform(
        map_data, fwhm=args.beam_arcmin * u.arcmin, rot=hp.Rotator(coord=to_coord)
    )

print("  Converting to NEST")
nested = hp.ud_grade(smoothed, args.nside, order_in="RING", order_out="NEST")

print("  Writing output...", flush=True)
hp.write_map(outfile, nested, nest=True)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from scipy.signal import fftconvolve

from ..mpi import MPI

from ..timing import function_timer

from .._libtoast import fod_autosums, fod_crosssums

from ..tod import flagged_running_average


def highpass_flagged_signal(sig, good, naverage):
    """Highpass-filter the signal to remove sub harmonic modes.

    Args:
        sig (array):  The signal.
        good (array):  Sample flags (zero == *BAD*)
        naverage (int):  The number of samples to average.

    Returns:
        (array):  The processed array.

    """
    # First fit and remove a linear trend.  Loss of power from this
    # filter is assumed negligible in the frequency bins of interest
    ngood = np.sum(good)
    if ngood == 0:
        raise RuntimeError("No valid samples")
    templates = np.vstack([np.ones(ngood), np.arange(good.size)[good]])
    invcov = np.dot(templates, templates.T)
    cov = np.linalg.inv(invcov)
    proj = np.dot(templates, sig[good])
    coeff = np.dot(cov, proj)
    sig[good] -= coeff[0] + coeff[1] * templates[1]
    # Then prewhiten the data.  This filter will be corrected in the
    # PSD estimates.
    trend = flagged_running_average(sig, good == 0, naverage)
    sig[good] -= trend[good]
    return sig


@function_timer
def autocov_psd(
    times,
    signal,
    flags,
    lagmax,
    stationary_period,
    fsample,
    comm=None,
    return_cov=False,
):
    """Compute the sample autocovariance.

    Compute the sample autocovariance function and Fourier transform it
    for a power spectral density. The resulting power spectral densities
    are distributed across the communicator as tuples of
    (start_time, stop_time, bin_frequency, bin_value)

    Args:
        times (float):  Signal time stamps.
        signal (float):  Regularly sampled signal vector.
        flags (float):  Signal quality flags.
        lagmax (int):  Largest sample separation to evaluate.
        stationary_period (float):  Length of a stationary interval in
            units of the times vector.
        fsample (float):  The sampling frequency in Hz
        comm (MPI.Comm):  The MPI communicator or None.
        return_cov (bool): Return also the covariance function

    Returns:
        (list):  List of local tuples of (start_time, stop_time, bin_frequency,
            bin_value)

    """
    return crosscov_psd(
        times, signal, None, flags, lagmax, stationary_period, fsample, comm, return_cov
    )


@function_timer
def crosscov_psd(
    times,
    signal1,
    signal2,
    flags,
    lagmax,
    stationary_period,
    fsample,
    comm=None,
    return_cov=False,
):
    """Compute the sample (cross)covariance.

    Compute the sample (cross)covariance function and Fourier transform it
    for a power spectral density. The resulting power spectral densities
    are distributed across the communicator as tuples of
    (start_time, stop_time, bin_frequency, bin_value)

    Args:
        times (float):  Signal time stamps.
        signal1 (float):  Regularly sampled signal vector.
        signal2 (float):  Regularly sampled signal vector or None.
        flags (float):  Signal quality flags.
        lagmax (int):  Largest sample separation to evaluate.
        stationary_period (float):  Length of a stationary interval in
            units of the times vector.
        fsample (float):  The sampling frequency in Hz
        comm (MPI.Comm):  The MPI communicator or None.
        return_cov (bool): Return also the covariance function

    Returns:
        (list):  List of local tuples of (start_time, stop_time, bin_frequency,
            bin_value)

    """
    rank = 0
    ntask = 1
    time_start = times[0]
    time_stop = times[-1]
    if comm is not None:
        rank = comm.rank
        ntask = comm.size
        time_start = comm.bcast(times[0], root=0)
        time_stop = comm.bcast(times[-1], root=ntask - 1)

    # We apply a prewhitening filter to the signal.  To accommodate the
    # quality flags, the filter is a moving average that only accounts
    # for the unflagged samples
    naverage = lagmax

    nreal = np.int(np.ceil((time_stop - time_start) / stationary_period))

    # Communicate lagmax samples from the beginning of the array
    # backwards in the MPI communicator

    nsamp = signal1.size

    if lagmax > nsamp:
        raise RuntimeError(
            "crosscov_psd: Communicating TOD beyond nearest neighbors is not "
            "implemented. Reduce lagmax or the size of the MPI communicator."
        )

    if rank != ntask - 1:
        nextend = lagmax
    else:
        nextend = 0

    extended_signal1 = np.zeros(nsamp + nextend, dtype=np.float64)
    if signal2 is not None:
        extended_signal2 = np.zeros(nsamp + nextend, dtype=np.float64)
    extended_flags = np.zeros(nsamp + nextend, dtype=np.bool)
    extended_times = np.zeros(nsamp + nextend, dtype=times.dtype)

    extended_signal1[:nsamp] = signal1
    if signal2 is not None:
        extended_signal2[:nsamp] = signal2
    extended_flags[:nsamp] = flags
    extended_times[:nsamp] = times

    if comm is not None:
        for evenodd in range(2):
            if rank % 2 == evenodd % 2:
                # Send
                if rank == 0:
                    continue
                comm.send(signal1[:lagmax], dest=rank - 1, tag=0)
                if signal2 is not None:
                    comm.send(signal1[:lagmax], dest=rank - 1, tag=3)
                comm.send(flags[:lagmax], dest=rank - 1, tag=1)
                comm.send(times[:lagmax], dest=rank - 1, tag=2)
            else:
                # Receive
                if rank == ntask - 1:
                    continue
                extended_signal1[-lagmax:] = comm.recv(source=rank + 1, tag=0)
                if signal2 is not None:
                    extended_signal1[-lagmax:] = comm.recv(source=rank + 1, tag=3)
                extended_flags[-lagmax:] = comm.recv(source=rank + 1, tag=1)
                extended_times[-lagmax:] = comm.recv(source=rank + 1, tag=2)

    realization = ((extended_times - time_start) / stationary_period).astype(np.int64)

    # Set flagged elements to zero

    extended_signal1[extended_flags != 0] = 0
    if signal2 is not None:
        extended_signal2[extended_flags != 0] = 0

    covs = {}

    for ireal in range(realization[0], realization[-1] + 1):
        # Evaluate the covariance
        realflg = realization == ireal
        good = extended_flags[realflg] == 0
        ngood = np.sum(good)
        if ngood == 0:
            continue
        sig1 = extended_signal1[realflg].copy()
        sig1 = highpass_flagged_signal(sig1, good, naverage)
        # High pass filter does not work at the ends
        ind = slice(naverage // 2, -naverage // 2)
        cov_hits = np.zeros(lagmax, dtype=np.int64)
        cov = np.zeros(lagmax, dtype=np.float64)
        if signal2 is None:
            fod_autosums(sig1[ind], good[ind].astype(np.uint8), lagmax, cov, cov_hits)
        else:
            sig2 = extended_signal2[realflg].copy()
            sig2 = highpass_flagged_signal(sig2, good, lagmax)
            fod_crosssums(
                sig1[ind], sig2[ind], good[ind].astype(np.uint8), lagmax, cov, cov_hits
            )
        covs[ireal] = (cov_hits, cov)

    # Collect the estimated covariance functions

    my_covs = {}
    nreal_task = np.int(np.ceil(nreal / ntask))

    if comm is None:
        for ireal in range(nreal):
            if ireal in covs:
                cov_hits, cov = covs[ireal]
            else:
                cov_hits = np.zeros(lagmax, dtype=np.int64)
                cov = np.zeros(lagmax, dtype=np.float64)
            my_covs[ireal] = (cov_hits, cov)
    else:
        for ireal in range(nreal):
            owner = ireal // nreal_task
            if ireal in covs:
                cov_hits, cov = covs[ireal]
            else:
                cov_hits = np.zeros(lagmax, dtype=np.int64)
                cov = np.zeros(lagmax, dtype=np.float64)

            cov_hits_total = np.zeros(lagmax, dtype=np.int64)
            cov_total = np.zeros(lagmax, dtype=np.float64)

            comm.Reduce(cov_hits, cov_hits_total, op=MPI.SUM, root=owner)
            comm.Reduce(cov, cov_total, op=MPI.SUM, root=owner)

            if rank == owner:
                my_covs[ireal] = (cov_hits_total, cov_total)

    # Now process the ones this task owns

    my_psds = []
    my_cov = []

    for ireal in my_covs.keys():
        cov_hits, cov = my_covs[ireal]
        good = cov_hits != 0
        cov[good] /= cov_hits[good]
        # Interpolate any empty bins
        if not np.all(good) and np.any(good):
            bad = cov_hits == 0
            # The last bins should be left empty
            i = cov.size - 1
            while cov_hits[i] == 0:
                cov[i] = 0
                bad[i] = False
                i -= 1
            nbad = np.sum(bad)
            if nbad > 0:
                good = np.logical_not(bad)
                lag = np.arange(lagmax)
                cov[bad] = np.interp(lag[bad], lag[good], cov[good])

        # Fourier transform for the PSD.  We symmetrize the sample
        # autocovariance so that the FFT is real-valued.  Notice that
        # we are not forcing the PSD to be positive:  each bin is a
        # noisy estimate of the true PSD.

        cov = np.hstack([cov, cov[:0:-1]])

        # w = np.roll(hamming(cov.size), -lagmax)
        # cov *= w

        psd = np.fft.rfft(cov).real
        psdfreq = np.fft.rfftfreq(len(cov), d=1 / fsample)

        # Post process the PSD estimate:
        #  1) Deconvolve the prewhitening (highpass) filter
        arg = 2 * np.pi * np.abs(psdfreq) * naverage / fsample
        tf = np.ones(lagmax)
        ind = arg != 0
        tf[ind] -= np.sin(arg[ind]) / arg[ind]
        psd[ind] /= tf[ind] ** 2
        #  2) Apply the Hann window to reduce unnecessary noise
        psd = np.convolve(psd, [0.25, 0.5, 0.25], mode="same")

        # Transfrom the corrected PSD back to get an unbiased
        # covariance function
        smooth_cov = np.fft.irfft(psd)
        my_cov.append((cov_hits, smooth_cov[:lagmax]))

        # Set the white noise PSD normalization to sigma**2 / fsample
        psd /= fsample

        tstart = time_start + ireal * stationary_period
        tstop = min(tstart + stationary_period, time_stop)

        my_psds.append((tstart, tstop, psdfreq, psd))

    if return_cov:
        return my_psds, my_cov
    else:
        return my_psds


def smooth_with_hits(hits, cov, wbin):
    """Smooth the covariance function.

    Smooth the covariance function, taking into account the number of hits in each bin.

    Args:
        hits (array_like):  The number of hits for each sample lag.
        cov (array_like):  The time domain covariance.
        wbin (int):  The number of samples per smoothing bin.

    Returns:
        (tuple): The (smoothed hits, smoothed covariance).

    """

    kernel = np.ones(wbin)
    smooth_hits = fftconvolve(hits, kernel, mode="same")
    smooth_cov = fftconvolve(cov * hits, kernel, mode="same")
    good = smooth_hits > 0
    smooth_cov[good] /= smooth_hits[good]

    return smooth_hits, smooth_cov



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.
"""Unit tests for the toast package.
"""

# If toast has not yet been imported, make sure we initialize MPI
from ..mpi import MPI

from .runner import test as run



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import numpy as np
import numpy.testing as nt

import healpy as hp

from .. import qarray as qa

from ..todmap import (
    TODHpixSpiral,
    OpPointingHpix,
    OpSimDipole,
    dipole,
    OpAccumDiag,
)

from ..map import DistPixels, covariance_invert, covariance_apply

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class OpSimDipoleTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # one detector per process and a single chunk.
        self.data = create_distdata(self.comm, obs_per_group=1)

        self.ndet = self.data.comm.group_size
        self.rate = 20.0

        # Create detectors with default properties
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet, samplerate=self.rate)

        # Pixelization
        self.nside = 64
        self.npix = 12 * self.nside ** 2

        # Samples per observation
        self.totsamp = self.npix

        # Dipole parameters
        self.solar_speed = 369.0
        gal_theta = np.deg2rad(90.0 - 48.05)
        gal_phi = np.deg2rad(264.31)
        z = self.solar_speed * np.cos(gal_theta)
        x = self.solar_speed * np.sin(gal_theta) * np.cos(gal_phi)
        y = self.solar_speed * np.sin(gal_theta) * np.sin(gal_phi)
        self.solar_vel = np.array([x, y, z])
        self.solar_quat = qa.from_vectors(np.array([0.0, 0.0, 1.0]), self.solar_vel)

        self.dip_check = 0.00335673

        self.dip_max_pix = hp.ang2pix(self.nside, gal_theta, gal_phi, nest=False)
        self.dip_min_pix = hp.ang2pix(
            self.nside, (np.pi - gal_theta), (np.pi + gal_phi), nest=False
        )

        # Populate the observations

        tod = TODHpixSpiral(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=self.data.comm.group_size,
            firsttime=0.0,
            rate=self.rate,
            nside=self.nside,
        )

        self.data.obs[0]["tod"] = tod

    def tearDown(self):
        del self.data

    def test_dipole_func(self):
        # Verify that we get the right magnitude if we are pointed at the
        # velocity maximum.
        dtod = dipole(self.solar_quat.reshape((1, 4)), solar=self.solar_vel)
        nt.assert_almost_equal(dtod, self.dip_check * np.ones_like(dtod))
        return

    def test_dipole_func_total(self):
        # Verify that we get the right magnitude if we are pointed at the
        # velocity maximum.

        quat = np.array(
            [
                [0.5213338, 0.47771442, -0.5213338, 0.47771442],
                [0.52143458, 0.47770023, -0.52123302, 0.4777286],
                [0.52153535, 0.47768602, -0.52113222, 0.47774277],
                [0.52163611, 0.4776718, -0.52103142, 0.47775692],
                [0.52173686, 0.47765757, -0.52093061, 0.47777106],
            ]
        )

        v_sat = np.array(
            [
                [1.82378638e-15, 2.97846918e01, 0.00000000e00],
                [-1.48252084e-07, 2.97846918e01, 0.00000000e00],
                [-2.96504176e-07, 2.97846918e01, 0.00000000e00],
                [-4.44756262e-07, 2.97846918e01, 0.00000000e00],
                [-5.93008348e-07, 2.97846918e01, 0.00000000e00],
            ]
        )

        v_sol = np.array([-25.7213418, -244.31203375, 275.33805175])

        dtod = dipole(quat, vel=v_sat, solar=v_sol, cmb=2.725)
        # computed with github.com/zonca/dipole
        expected = np.array(
            [0.00196249, 0.00196203, 0.00196157, 0.00196111, 0.00196065]
        )
        nt.assert_allclose(dtod, expected, rtol=1e-5)
        return

    def test_sim(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.nside, nest=False, mode="I")
        pointing.exec(self.data)

        # generate timestreams
        op = OpSimDipole(mode="solar", coord="G")
        op.exec(self.data)

        # make a binned map

        # construct distributed maps to store the covariance,
        # noise weighted map, and hits

        invnpp = DistPixels(
            self.data,
            nnz=1,
            dtype=np.float64,
            nest=False,
        )
        invnpp.data.fill(0.0)

        zmap = DistPixels(
            self.data,
            nnz=1,
            dtype=np.float64,
            nest=False,
        )
        zmap.data.fill(0.0)

        hits = DistPixels(
            self.data,
            nnz=1,
            dtype=np.int64,
            nest=False,
        )
        hits.data.fill(0)

        # accumulate the inverse covariance and noise weighted map.
        # Use detector weights based on the analytic NET.

        tod = self.data.obs[0]["tod"]

        detweights = {}
        for d in tod.local_dets:
            detweights[d] = 1.0

        build_invnpp = OpAccumDiag(
            detweights=detweights, invnpp=invnpp, hits=hits, zmap=zmap, name="dipole"
        )
        build_invnpp.exec(self.data)

        invnpp.allreduce()
        hits.allreduce()
        zmap.allreduce()

        hits.write_healpix_fits(os.path.join(self.outdir, "hits.fits"))
        invnpp.write_healpix_fits(os.path.join(self.outdir, "invnpp.fits"))
        zmap.write_healpix_fits(os.path.join(self.outdir, "zmap.fits"))

        # invert it
        covariance_invert(invnpp, 1.0e-3)

        invnpp.write_healpix_fits(os.path.join(self.outdir, "npp.fits"))

        # compute the binned map, N_pp x Z

        covariance_apply(invnpp, zmap)
        zmap.write_healpix_fits(os.path.join(self.outdir, "binned.fits"))

        if rank == 0:
            import matplotlib.pyplot as plt

            mapfile = os.path.join(self.outdir, "hits.fits")
            data = hp.read_map(mapfile, nest=False)
            nt.assert_almost_equal(
                data, self.data.comm.ngroups * self.ndet * np.ones_like(data)
            )

            outfile = "{}.png".format(mapfile)
            hp.mollview(data, xsize=1600, nest=False)
            plt.savefig(outfile)
            plt.close()

            mapfile = os.path.join(self.outdir, "binned.fits")
            data = hp.read_map(mapfile, nest=False)

            # verify that the extrema are in the correct location
            # and have the correct value.

            minmap = np.min(data)
            maxmap = np.max(data)
            nt.assert_almost_equal(maxmap, self.dip_check, decimal=5)
            nt.assert_almost_equal(minmap, -self.dip_check, decimal=5)

            minloc = np.argmin(data)
            maxloc = np.argmax(data)
            nt.assert_equal(minloc, self.dip_min_pix)
            nt.assert_equal(maxloc, self.dip_max_pix)

            outfile = "{}.png".format(mapfile)
            hp.mollview(data, xsize=1600, nest=False)
            plt.savefig(outfile)
            plt.close()
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os
import numpy as np

import h5py

from ..utils import Environment, Logger

from ..timing import Timer, function_timer, GlobalTimers

from ..rng import random as toast_rng

from ..mpi import MPI, MPIShared

from .._libtoast import (
    atm_sim_compute_slice,
    atm_sim_observe,
    atm_sim_compress_flag_hits_rank,
    atm_sim_compress_flag_extend_rank,
    atm_sim_kolmogorov_init_rank,
)

available_utils = None
if available_utils is None:
    available_utils = True
    try:
        from .._libtoast import (
            atm_absorption_coefficient,
            atm_absorption_coefficient_vec,
            atm_atmospheric_loading,
            atm_atmospheric_loading_vec,
        )
    except ImportError:
        available_utils = False


class AtmSim(object):
    """Class representing a single atmosphere simulation.

    This simulation consists of a particular realization of a "slab" of the
    atmosphere that moves with a constant wind speed and can be observed by
    individual detectors.

    Args:
        azmin (float):  The minimum of the azimuth range.
        azmax (float):  The maximum of the azimuth range.
        elmin (float):  The minimum of the elevation range.
        elmax (float):  The maximum of the elevation range.
        tmin (float):  The minimum of the time range.
        tmax (float):  The maximum of the time range.
        lmin_center (float):  Center point of the distribution of the dissipation
            scale of the Kolmogorov turbulence.
        lmin_sigma (float):  Width of the distribution of the dissipation
            scale of the Kolmogorov turbulence.
        lmax_center (float):  Center point of the distribution of the injection
            scale of the Kolmogorov turbulence.
        lmax_sigma (float):  Width of the distribution of the injection
            scale of the Kolmogorov turbulence.
        w_center (float):  Center point of the distribution of wind speed (m/s).
        w_sigma (float):  Width of the distribution of wind speed.
        wdir_center (float):  Center point of the distribution of wind direction
            (radians).
        wdir_sigma (float):  Width of the distribution of wind direction.
        z0_center (float):  Center point of the distribution of the water vapor (m).
        z0_sigma (float):  Width of the distribution of the water vapor.
        T0_center (float):  Center point of the distribution of ground temperature
            (Kelvin).
        T0_sigma (float):  Width of the distribution of ground temperature.
        zatm (float):  Atmosphere extent for temperature profile.
        zmax (float):  Water vaport extent for integration.
        xstep (float):  Size of volume element in the X direction.
        ystep (float):  Size of volume element in the Y direction.
        zstep (float):  Size of volume element in the Z direction.
        nelem_sim_max (int):  Size of the simulation slices.
        comm (mpi4py.MPI.Comm):  The MPI communicator or None.
        key1 (uint64):  Streamed RNG key 1.
        key2 (uint64):  Streamed RNG key 2.
        counterval1 (uint64):  Streamed RNG counter 1.
        counterval2 (uint64):  Streamed RNG counter 2.
        cachedir (str):  The location of the cached simulation.
        rmin (float):  Minimum line of sight observing distance.
        rmax (float):  Maximum line of sight observing distance.
        write_debug (bool): If true, write out intermediate text files for debugging.

    """

    def __init__(
        self,
        azmin,
        azmax,
        elmin,
        elmax,
        tmin,
        tmax,
        lmin_center,
        lmin_sigma,
        lmax_center,
        lmax_sigma,
        w_center,
        w_sigma,
        wdir_center,
        wdir_sigma,
        z0_center,
        z0_sigma,
        T0_center,
        T0_sigma,
        zatm,
        zmax,
        xstep,
        ystep,
        zstep,
        nelem_sim_max,
        comm,
        key1,
        key2,
        counterval1,
        counterval2,
        cachedir,
        rmin,
        rmax,
        write_debug=False,
    ):
        self._azmin = azmin
        self._azmax = azmax
        self._elmin = elmin
        self._elmax = elmax
        self._tmin = tmin
        self._tmax = tmax
        self._lmin_center = lmin_center
        self._lmin_sigma = lmin_sigma
        self._lmax_center = lmax_center
        self._lmax_sigma = lmax_sigma
        self._w_center = w_center
        self._w_sigma = w_sigma
        self._wdir_center = wdir_center
        self._wdir_sigma = wdir_sigma
        self._z0_center = z0_center
        self._z0_sigma = z0_sigma
        self._T0_center = T0_center
        self._T0_sigma = T0_sigma
        self._zatm = zatm
        self._zmax = zmax
        self._xstep = xstep
        self._ystep = ystep
        self._zstep = zstep
        self._nelem_sim_max = nelem_sim_max
        self._comm = comm
        self._key1 = key1
        self._key2 = key2
        self._counter1start = counterval1
        self._counter2start = counterval2
        self._cachedir = cachedir
        self._rmin = rmin
        self._rmax = rmax
        self._write_debug = write_debug

        self._counter1 = self._counter1start
        self._counter2 = self._counter2start

        self._corrlim = 1e-3

        self._ntask = 1
        self._rank = 0
        if self._comm is not None:
            self._ntask = self._comm.size
            self._rank = self._comm.rank

        env = Environment.get()
        self._nthread = env.max_threads()

        log = Logger.get()

        if self._rank == 0:
            log.debug(
                "AtmSim constructed with {} processes, {} threads per process".format(
                    self._ntask, self._nthread
                )
            )

        if azmin >= azmax:
            raise RuntimeError("AtmSim: azmin >= azmax")

        if elmin < 0:
            raise RuntimeError("AtmSim: elmin < 0")

        if elmax > 0.5 * np.pi:
            raise RuntimeError("AtmSim: elmax > pi/2")

        if elmin > elmax:
            raise RuntimeError("AtmSim: elmin >= elmax")

        if tmin >= tmax:
            raise RuntimeError("AtmSim: tmin >= tmax")

        if lmin_center > lmax_center:
            raise RuntimeError("AtmSim: lmin_center >= lmax_center")

        self._delta_az = self._azmax - self._azmin
        self._delta_el = self._elmax - self._elmin
        self._delta_t = self._tmax - self._tmin

        self._az0 = self._azmin + self._delta_az / 2
        self._el0 = self._elmin + self._delta_el / 2
        self._sinel0 = np.sin(self._el0)
        self._cosel0 = np.cos(self._el0)

        self._xxstep = self._xstep * self._cosel0 - self._zstep * self._sinel0
        self._yystep = self._ystep
        self._zzstep = self._xstep * self._sinel0 + self._zstep * self._cosel0

        if self._rank == 0:
            msg = "\nInput parameters:\n"
            msg += "             az = [{} - {}] ({} degrees)\n".format(
                np.degrees(self._azmin),
                np.degrees(self._azmax),
                np.degrees(self._delta_az),
            )
            msg += "             el = [{} - {}] ({} degrees)\n".format(
                np.degrees(self._elmin),
                np.degrees(self._elmax),
                np.degrees(self._delta_el),
            )
            msg += "              t = [{} - {}] ({} s)\n".format(
                self._tmin, self._tmax, self._delta_t
            )
            msg += "           lmin = {} +- {} m\n".format(
                self._lmin_center, self._lmin_sigma
            )
            msg += "           lmax = {} +- {} m\n".format(
                self._lmax_center, self._lmax_sigma
            )
            msg += "              w = {} +- {} m/s\n".format(
                self._w_center, self._w_sigma
            )
            msg += "           wdir = {} +- {} degrees\n".format(
                np.degrees(self._wdir_center), np.degrees(self._wdir_sigma)
            )
            msg += "             z0 = {} +- {} m\n".format(
                self._z0_center, self._z0_sigma
            )
            msg += "             T0 = {} +- {} K\n".format(
                self._T0_center, self._T0_sigma
            )
            msg += "           zatm = {} m\n".format(self._zatm)
            msg += "           zmax = {} m\n".format(self._zmax)
            msg += "       scan frame:\n"
            msg += "          xstep = {} m\n".format(self._xstep)
            msg += "          ystep = {} m\n".format(self._ystep)
            msg += "          zstep = {} m\n".format(self._zstep)
            msg += "  nelem_sim_max = {}\n".format(self._nelem_sim_max)
            msg += "        corrlim = {}\n".format(self._corrlim)
            msg += "           rmin = {} m\n".format(self._rmin)
            msg += "           rmax = {} m\n".format(self._rmax)
            log.debug(msg)

        self._compressed_index = None
        self._full_index = None
        self._realization = None
        self._cached = False
        self._nelem = None

    @function_timer
    def simulate(self, use_cache=False, smooth=False):
        """Perform the simulation.

        Args:
            use_cache (bool):  If True, load / save from / to cachedir.
            smooth (bool):  If True, apply smoothing.

        Returns:
            (int):  A status value (zero == good).

        """
        if use_cache:
            if self._cachedir is None:
                raise RuntimeError("Cannot use the cache if cachedir is not set")
            self.load_realization()

        if self._cached:
            return 0

        log = Logger.get()

        self.draw()

        self.get_volume()

        self.compress_volume()

        if self._rank == 0:
            log.debug("Resizing realization to {}".format(self._nelem))

        timer = Timer()
        timer.start()

        # This memory is already zeroed on construction.
        self._realization = MPIShared((self._nelem,), np.float64, self._comm)

        timer.stop()
        if self._rank == 0:
            log.debug(
                "Allocated shared realization buffer in {} s".format(timer.seconds())
            )
        timer.clear()
        timer.start()

        ind_start = 0
        ind_stop = 0
        slice = 0

        # Simulate the atmosphere in independent slices, each slice
        # assigned to one process.

        slice_starts = list()
        slice_stops = list()

        # Set a per-process global timer to check for load imbalance
        gt = GlobalTimers.get()
        gt.start("AtmSim compute slices")

        while True:
            ind_start, ind_stop = self._get_slice(ind_start, ind_stop)
            slice_starts.append(ind_start)
            slice_stops.append(ind_stop)

            if slice % self._ntask == self._rank:
                atm_sim_compute_slice(
                    ind_start,
                    ind_stop,
                    self._rmin_kolmo,
                    self._rmax_kolmo,
                    self._kolmo_x,
                    self._kolmo_y,
                    self._rcorr,
                    self._xstart,
                    self._ystart,
                    self._zstart,
                    self._xstep,
                    self._ystep,
                    self._zstep,
                    self._xstride,
                    self._ystride,
                    self._zstride,
                    self._z0,
                    self._cosel0,
                    self._sinel0,
                    self._full_index.data,
                    smooth,
                    self._xxstep,
                    self._zzstep,
                    self._rank,
                    self._key1,
                    self._key2,
                    self._counter1,
                    self._counter2,
                    self._realization.data,
                )

            # Advance the RNG counter on all processes
            self._counter2 += ind_stop - ind_start

            if ind_stop == self._nelem:
                break

            slice += 1

        slice_starts = np.array(slice_starts, dtype=np.int32)
        slice_stops = np.array(slice_stops, dtype=np.int32)

        gt.stop("AtmSim compute slices")
        if self._comm is not None:
            self._comm.barrier()

        timer.stop()
        if self._rank == 0:
            log.debug("Simulated all slices in {} s".format(timer.seconds()))
        timer.clear()
        timer.start()

        # Gather the slices.  Every process has written their assigned slices to their
        # node-local shared memory buffer.  However, these slices do not exist on the
        # other nodes.  For each slice, the assigned process copies the slice to a
        # buffer and then uses the MPIShared.set() method to replicate to shared
        # memory on all nodes.

        for slice in range(len(slice_starts)):
            ind_start = slice_starts[slice]
            ind_stop = slice_stops[slice]
            nind = ind_stop - ind_start
            root = slice % self._ntask

            tempvec = None
            if self._rank == root:
                # This process has the slice, copy to a buffer
                tempvec = np.array(self._realization.data[ind_start:ind_stop])

            # Set the slice across all nodes
            self._realization.set(tempvec, (ind_start,), fromrank=root)

        timer.stop()
        if self._rank == 0:
            log.debug("Distributed slices to all nodes in {} s".format(timer.seconds()))
        timer.clear()
        timer.start()

        # self.smooth()

        self._cached = True
        if use_cache:
            self.save_realization()

        timer.stop()
        if use_cache and self._rank == 0:
            log.debug("Saved realization in {} s".format(timer.seconds()))

        return 0

    @function_timer
    def observe(self, times, az, el, tod, fixed_r=-1):
        """Observe the atmosphere with a detector.

        The timestamps and Azimuth / Elevation pointing are provided.  The TOD
        buffer is filled with the integrated atmospheric signal.

        For each sample, integrate along the line of sight by summing the
        atmosphere values. See Church (1995) Section 2.2, first equation.
        We omit the optical depth factor which is close to unity.

        Args:
            times (array_like):  Detector timestamps.
            az (array like):  Azimuth values.
            el (array_like):  Elevation values.
            tod (array_like):  The output buffer to fill.
            fixed_r (float):  If greater than zero, use this single radial value.

        Returns:
            (int):  A status value (zero == good).

        """
        if not self._cached:
            raise RuntimeError("There is no cached observation to observe")

        log = Logger.get()
        timer = Timer()
        timer.start()

        nsamp = len(times)

        status = 0

        status = atm_sim_observe(
            times,
            az,
            el,
            tod,
            self._T0,
            self._azmin,
            self._azmax,
            self._elmin,
            self._elmax,
            self._tmin,
            self._tmax,
            self._rmin,
            self._rmax,
            fixed_r,
            self._zatm,
            self._zmax,
            self._wx,
            self._wy,
            self._wz,
            self._xstep,
            self._ystep,
            self._zstep,
            self._xstart,
            self._delta_x,
            self._ystart,
            self._delta_y,
            self._zstart,
            self._delta_z,
            self._maxdist,
            self._nx,
            self._ny,
            self._nz,
            self._xstride,
            self._ystride,
            self._zstride,
            self._compressed_index.data,
            self._full_index.data,
            self._realization.data,
        )

        timer.stop()

        if self._rank == 0:
            if fixed_r > 0:
                log.debug(
                    "Observed {} samples at r = {} in {} s".format(
                        nsamp, fixed_r, timer.seconds()
                    )
                )
            else:
                log.debug("Observed {} samples in {} s".format(nsamp, timer.seconds()))

        if status != 0:
            log.error("Observing {} samples failed with error {}".format(nsamp, status))
        return status

    def _get_slice(self, ind_start, ind_stop):
        """Identify a manageable slice of compressed indices to simulate next."""
        log = Logger.get()

        # Move element counter to the end of the most recent simulated slice
        ind_start = ind_stop

        # TK: The original code multiplied a double and a long and implicitly cast
        # back to a long.  implementing that here, but should verify that it is
        # correct.
        xstrideinv = 1.0 / self._xstride
        ix_start = int(self._full_index[ind_start] * xstrideinv)
        ix1 = ix_start
        ix2 = 0

        while True:
            # Advance element counter by one layer of elements
            ix2 = ix1
            while ix1 == ix2:
                ind_stop += 1
                if ind_stop == self._nelem:
                    break
                ix2 = int(self._full_index[ind_stop] * xstrideinv)
            # Check if there are no more elements
            if ind_stop == self._nelem:
                break

            # Check if we have enough to meet the minimum number of elements
            if ind_stop - ind_start >= self._nelem_sim_max:
                break

            # Check if we have enough layers
            # nlayer_sim_max = 10
            # if ix2 - ix_start >= nlayer_sim_max:
            #     break
            ix1 = ix2

        if self._rank == 0:
            log.debug(
                "X-slice: {} -- {} ({} {} m layers) m out of {} m indices {} -- {} ({}) out of {}".format(
                    ix_start * self._xstep,
                    ix2 * self._xstep,
                    ix2 - ix_start,
                    self._xstep,
                    self._nx * self._xstep,
                    ind_start,
                    ind_stop,
                    ind_stop - ind_start,
                    self._nelem,
                )
            )
        return (ind_start, ind_stop)

    def draw(self):
        # Draw gaussian variates to use in drawing the simulation
        # parameters.
        log = Logger.get()

        nrand = 10000

        randn = np.array(
            toast_rng(
                nrand,
                key=(self._key1, self._key2),
                counter=(self._counter1, self._counter2),
                sampler="gaussian",
                threads=False,
            ),
            dtype=np.float64,
        )

        self._counter2 += nrand

        irand = 0

        # The random numbers above are identical on all processes, so every process
        # just computes the same parameters locally.

        self._lmin = 0
        self._lmax = 0
        self._w = -1
        self._wdir = 0
        self._z0 = 0
        self._T0 = 0

        while self._lmin >= self._lmax:
            self._lmin = 0
            self._lmax = 0
            while (self._lmin <= 0) and (irand < nrand - 1):
                self._lmin = self._lmin_center + randn[irand] * self._lmin_sigma
                irand += 1
            while (self._lmax <= 0) and (irand < nrand - 1):
                self._lmax = self._lmax_center + randn[irand] * self._lmax_sigma
                irand += 1

        while (self._w < 0) and (irand < nrand - 1):
            self._w = self._w_center + randn[irand] * self._w_sigma
            irand += 1

        self._wdir = np.fmod(self._wdir_center + randn[irand] * self._wdir_sigma, np.pi)
        irand += 1

        while (self._z0 <= 0) and (irand < nrand):
            self._z0 = self._z0_center + randn[irand] * self._z0_sigma
            irand += 1
        while (self._T0 <= 0) and (irand < nrand):
            self._T0 = self._T0_center + randn[irand] * self._T0_sigma
            irand += 1

        if irand == nrand:
            raise RuntimeError(
                "Failed to draw parameters so satisfy boundary conditions"
            )

        # Wind is parallel to surface. Rotate to a frame where the scan
        # is across the X-axis.

        # self._w *= 0.5  # DEBUG
        # self._wdir += np.pi  # DEBUG

        eastward_wind = self._w * np.cos(self._wdir)
        northward_wind = self._w * np.sin(self._wdir)

        angle = self._az0 - 0.5 * np.pi
        wx_h = eastward_wind * np.cos(angle) - northward_wind * np.sin(angle)
        self._wy = eastward_wind * np.sin(angle) + northward_wind * np.cos(angle)

        self._wx = wx_h * self._cosel0
        self._wz = -wx_h * self._sinel0

        # Inverse the wind direction so we can apply it to the
        # telescope position

        self._wx *= -1
        self._wy *= -1
        self._wz *= -1

        if self._rank == 0:
            msg = "\nAtmospheric realization parameters:\n"
            msg += "           lmin = {} m\n".format(self._lmin)
            msg += "           lmax = {} m\n".format(self._lmax)
            msg += "              w = {} m/s\n".format(self._w)
            msg += "  eastward wind = {} m/s\n".format(eastward_wind)
            msg += " northward wind = {} m/s\n".format(northward_wind)
            msg += "            az0 = {} degrees\n".format(np.degrees(self._az0))
            msg += "            el0 = {} degrees\n".format(np.degrees(self._el0))
            msg += "             wx = {} m/s\n".format(self._wx)
            msg += "             wy = {} m/s\n".format(self._wy)
            msg += "             wz = {} m/s\n".format(self._wz)
            msg += "           wdir = {} degrees\n".format(np.degrees(self._wdir))
            msg += "             z0 = {} m\n".format(self._z0)
            msg += "             T0 = {} K\n".format(self._T0)
            log.debug(msg)
        return

    def get_volume(self):
        """Compute the volume."""
        log = Logger.get()

        # Trim zmax if rmax sets a more stringent limit
        zmax_test = self._rmax * np.sin(self._elmax)
        if self._zmax > zmax_test:
            self._zmax = zmax_test

        # Horizontal volume

        delta_z_h = self._zmax

        # Maximum distance observed through the simulated volume

        self._maxdist = delta_z_h / self._sinel0

        # Volume length
        delta_x_h = self._maxdist * np.cos(self._elmin)

        # double x, y, z, xx, zz, r, rproj, z_min, z_max;

        r = self._maxdist

        z = r * np.sin(self._elmin)
        rproj = r * np.cos(self._elmin)
        x = rproj * np.cos(0)
        z_min = -x * self._sinel0 + z * self._cosel0

        z = r * np.sin(self._elmax)
        rproj = r * np.cos(self._elmax)
        x = rproj * np.cos(self._delta_az / 2)
        z_max = -x * self._sinel0 + z * self._cosel0

        # Cone width
        rproj = r * np.cos(self._elmin)
        delta_y_cone = 0
        if self._delta_az > np.pi:
            delta_y_cone = 2 * rproj
        else:
            delta_y_cone = 2 * rproj * np.cos(0.5 * (np.pi - self._delta_az))

        # Cone height
        delta_z_cone = z_max - z_min

        # Rotate to observation plane
        self._delta_x = self._maxdist
        self._delta_z = delta_z_cone

        # Wind effects

        wdx = np.absolute(self._wx) * self._delta_t
        wdy = np.absolute(self._wy) * self._delta_t
        wdz = np.absolute(self._wz) * self._delta_t
        self._delta_x += wdx
        self._delta_y = delta_y_cone + wdy
        self._delta_z += wdz

        # Margin for interpolation

        self._delta_x += self._xstep
        self._delta_y += 2 * self._ystep
        self._delta_z += 2 * self._zstep

        # Translate the volume to allow for wind.  Telescope sits
        # at (0, 0, 0) at t=0

        if self._wx < 0:
            self._xstart = -wdx
        else:
            self._xstart = 0

        if self._wy < 0:
            self._ystart = -0.5 * delta_y_cone - wdy - self._ystep
        else:
            self._ystart = -0.5 * delta_y_cone - self._ystep

        if self._wz < 0:
            self._zstart = z_min - wdz - self._zstep
        else:
            self._zstart = z_min - self._zstep

        # Grid points

        # TK: the original code had an implicity conversion from double to long,
        # which is replicated here.  Is this the desired behavior?
        self._nx = int(self._delta_x / self._xstep) + 1
        self._ny = int(self._delta_y / self._ystep) + 1
        self._nz = int(self._delta_z / self._zstep) + 1
        self._nn = self._nx * self._ny * self._nz

        # 1D storage of the 3D volume elements

        self._zstride = 1
        self._ystride = self._zstride * self._nz
        self._xstride = self._ystride * self._ny

        if self._rank == 0:
            msg = "Simulation volume:\n"
            msg += "       delta_x = {} m\n".format(self._delta_x)
            msg += "       delta_y = {} m\n".format(self._delta_y)
            msg += "       delta_z = {} m\n".format(self._delta_z)
            msg += "Observation cone along the X-axis:\n"
            msg += "  delta_y_cone = {} m\n".format(delta_y_cone)
            msg += "  delta_z_cone = {} m\n".format(delta_z_cone)
            msg += "        xstart = {} m\n".format(self._xstart)
            msg += "        ystart = {} m\n".format(self._ystart)
            msg += "        zstart = {} m\n".format(self._zstart)
            msg += "       maxdist = {} m\n".format(self._maxdist)
            msg += "            nx = {}\n".format(self._nx)
            msg += "            ny = {}\n".format(self._ny)
            msg += "            nz = {}\n".format(self._nz)
            msg += "            nn = {}\n".format(self._nn)
            log.debug(msg)

        self._initialize_kolmogorov()
        return

    @function_timer
    def compress_volume(self):
        """Establish a mapping between full and observed volume indices."""
        log = Logger.get()

        if self._rank == 0:
            log.debug("Compressing volume, N = {}".format(self._nn))

        timer = Timer()
        timer.start()

        self._compressed_index = MPIShared((self._nn,), np.int64, self._comm)
        self._full_index = MPIShared((self._nn,), np.int64, self._comm)

        timer.stop()
        if self._rank == 0:
            log.debug(
                "Allocated shared full and compressed indices in {} s".format(
                    timer.seconds()
                )
            )
        timer.clear()
        timer.start()

        hit = np.zeros(self._nn, dtype=np.uint8)

        # Start by flagging all elements that are hit

        atm_sim_compress_flag_hits_rank(
            hit,
            self._ntask,
            self._rank,
            self._nx,
            self._ny,
            self._nz,
            self._xstart,
            self._ystart,
            self._zstart,
            self._delta_t,
            self._delta_az,
            self._elmin,
            self._elmax,
            self._wx,
            self._wy,
            self._wz,
            self._xstep,
            self._ystep,
            self._zstep,
            self._xstride,
            self._ystride,
            self._zstride,
            self._maxdist,
            self._cosel0,
            self._sinel0,
        )

        timer.stop()
        if self._rank == 0:
            log.debug("Flagged hits in {} s".format(timer.seconds()))
        timer.clear()
        timer.start()

        if self._comm is not None:
            self._comm.Allreduce(MPI.IN_PLACE, [hit, MPI.UNSIGNED_CHAR], MPI.LOR)

        timer.stop()
        if self._rank == 0:
            log.debug("Reduced hits in {} s".format(timer.seconds()))
        timer.clear()
        timer.start()

        # For extra margin, flag all the neighbors of the hit elements

        hit2 = np.array(hit)

        atm_sim_compress_flag_extend_rank(
            hit,
            hit2,
            self._ntask,
            self._rank,
            self._nx,
            self._ny,
            self._nz,
            self._xstride,
            self._ystride,
            self._zstride,
        )

        timer.stop()
        if self._rank == 0:
            log.debug("Extended flags in {} s".format(timer.seconds()))
        timer.clear()
        timer.start()

        del hit2

        if self._comm is not None:
            self._comm.Allreduce(MPI.IN_PLACE, [hit, MPI.UNSIGNED_CHAR], MPI.LOR)

        timer.stop()
        if self._rank == 0:
            log.debug("Reduced extended hits in {} s".format(timer.seconds()))
        timer.clear()
        timer.start()

        # Create the mappings between the compressed and full indices.  The hit
        # vector is duplicated across all processes and we are modifying node shared
        # memory.  Just one process per node manipulates the copy on each node.

        if (self._full_index._nodecomm is None) or (
            self._full_index._nodecomm.rank == 0
        ):
            # we are the root process on this node.
            i = 0
            for ifull in range(self._nn):
                if hit[ifull]:
                    self._full_index.data[i] = ifull
                    self._compressed_index.data[ifull] = i
                    i += 1
            self._nelem = i
        if self._full_index._nodecomm is not None:
            self._nelem = self._full_index._nodecomm.bcast(self._nelem, root=0)

        timer.stop()
        if self._rank == 0:
            log.debug("Created compression table in {} s".format(timer.seconds()))
        timer.clear()
        timer.start()

        del hit

        # Shrink the full index to what is needed.  Only one process per node copies
        # existing data.

        new_full = MPIShared((self._nelem,), np.int64, self._comm)
        if (self._full_index._nodecomm is None) or (
            self._full_index._nodecomm.rank == 0
        ):
            new_full.data[:] = self._full_index.data[: self._nelem]

        if self._comm is not None:
            self._comm.barrier()

        self._full_index.close()
        del self._full_index

        self._full_index = new_full

        if self._rank == 0:
            msg = "\nVolume compressed:\n"
            msg += "  {} / {} ({} %) elements needed for the simulation\n".format(
                self._nelem, self._nn, (self._nelem * 100 / self._nn)
            )
            msg += "  nx = {} ny = {} nz = {}\n".format(self._nx, self._ny, self._nz)
            msg += "  wx = {} wy = {} wz = {}\n".format(self._wx, self._wy, self._wz)
            log.debug(msg)
        if self._nelem == 0:
            raise RuntimeError("No elements in the observation cone.")
        return

    def smooth(self):
        pass

    def _meta_keys(self):
        """Helper function to return the list of member variables to read / write."""
        return list(
            [
                ("nn", int),
                ("nelem", int),
                ("nx", int),
                ("ny", int),
                ("nz", int),
                ("xstride", int),
                ("ystride", int),
                ("zstride", int),
                ("delta_x", float),
                ("delta_y", float),
                ("delta_z", float),
                ("xstart", float),
                ("ystart", float),
                ("zstart", float),
                ("maxdist", float),
                ("wx", float),
                ("wy", float),
                ("wz", float),
                ("lmin", float),
                ("lmax", float),
                ("w", float),
                ("wdir", float),
                ("z0", float),
                ("T0", float),
            ]
        )

    @function_timer
    def save_realization(self):
        if (
            (self._realization is None)
            or (self._full_index is None)
            or (self._compressed_index is None)
        ):
            raise RuntimeError("Cannot save realization: not loaded or generated")

        if self._rank == 0:

            log = Logger.get()

            rname = "{}_{}_{}_{}".format(
                self._key1, self._key2, self._counter1start, self._counter2start
            )
            outfile = os.path.join(self._cachedir, "{}.h5".format(rname))
            tmpfile = "{}.tmp".format(outfile)

            hf = h5py.File(tmpfile, "w")

            # Store metadata as attributes of the root group
            meta = hf.attrs
            for k, tp in self._meta_keys():
                meta.create(k, getattr(self, "_{}".format(k)))

            log.debug("Saved metadata for {}".format(rname))

            realiz = hf.create_dataset("realization", data=self._realization.data)
            log.debug("Saved realization for {}".format(rname))

            full = hf.create_dataset("full_index", data=self._full_index.data)
            log.debug("Saved full index for {}".format(rname))

            comp = hf.create_dataset(
                "compressed_index", data=self._compressed_index.data
            )
            log.debug("Saved compressed index for {}".format(rname))
            hf.flush()
            hf.close()

            # Move file into place
            os.rename(tmpfile, outfile)

        if self._comm is not None:
            self._comm.barrier()
        return

    @function_timer
    def load_realization(self):
        rname = None
        cachefile = None
        found = False
        if self._rank == 0:
            rname = "{}_{}_{}_{}".format(
                self._key1, self._key2, self._counter1start, self._counter2start
            )
            cachefile = os.path.join(self._cachedir, "{}.h5".format(rname))
            if os.path.isfile(cachefile):
                found = True
        if self._comm is not None:
            found = self._comm.bcast(found, root=0)

        if not found:
            return

        if self._realization is not None:
            del self._realization
        if self._full_index is not None:
            del self._full_index
        if self._compressed_index is not None:
            del self._compressed_index

        mdata = dict()
        hf = None
        if self._rank == 0:
            log = Logger.get()

            hf = h5py.File(cachefile, "r")
            # Read metadata
            meta = hf.attrs
            for k, tp in self._meta_keys():
                # Copy the metadata value into a dictionary, casting to correct type
                mdata[k] = tp(meta[k])
            log.debug("Loaded metadata for {}".format(rname))

        # Broadcast the metadata
        if self._comm is not None:
            mdata = self._comm.bcast(mdata, root=0)

        # Copy the metadata into class instance member variables
        for k, tp in self._meta_keys():
            setattr(self, "_{}".format(k), mdata[k])

        # Allocate the shared memory buffers
        self._realization = MPIShared((self._nelem,), np.float64, self._comm)
        self._compressed_index = MPIShared((self._nn,), np.int64, self._comm)
        self._full_index = MPIShared((self._nelem,), np.int64, self._comm)

        # Read and set shared memory
        buffer = None
        if self._rank == 0:
            buffer = np.zeros(self._nelem, dtype=np.float64)
            hf["realization"].read_direct(buffer)
            log.debug("Loaded realization for {}".format(rname))
        self._realization.set(buffer, (0,), fromrank=0)

        if self._rank == 0:
            buffer = np.zeros(self._nelem, dtype=np.int64)
            hf["full_index"].read_direct(buffer)
            log.debug("Loaded full index for {}".format(rname))
        self._full_index.set(buffer, (0,), fromrank=0)

        if self._rank == 0:
            buffer = np.zeros(self._nn, dtype=np.int64)
            hf["compressed_index"].read_direct(buffer)
            log.debug("Loaded compressed index for {}".format(rname))
        self._compressed_index.set(buffer, (0,), fromrank=0)

        if self._comm is not None:
            self._comm.barrier()

        self._cached = True
        return

    @function_timer
    def _initialize_kolmogorov(self):
        log = Logger.get()
        timer = Timer()
        timer.start()

        # Size of interpolation grid
        self._nr = 1000

        self._rmin_kolmo = 0.0
        diag = np.sqrt(self._delta_x ** 2 + self._delta_y ** 2)
        self._rmax_kolmo = np.sqrt(diag ** 2 + self._delta_z ** 2) * 1.01

        self._rstep = (self._rmax_kolmo - self._rmin_kolmo) / (self._nr - 1)

        self._kolmo_x, self._kolmo_y = atm_sim_kolmogorov_init_rank(
            self._nr,
            self._rmin_kolmo,
            self._rmax_kolmo,
            self._rstep,
            self._lmin,
            self._lmax,
            self._ntask,
            self._rank,
        )

        if self._comm is not None:
            self._comm.Allreduce(MPI.IN_PLACE, [self._kolmo_y, MPI.DOUBLE], MPI.SUM)

        # Normalize

        norm = 1.0 / self._kolmo_y[0]
        self._kolmo_y *= norm

        if (self._rank == 0) and self._write_debug:
            with open("kolmogorov.txt", "w") as f:
                for ir in range(self._nr):
                    f.write(
                        "{:.15e} {:.15e}\n".format(self._kolmo_x[ir], self._kolmo_y[ir])
                    )

        # Measure the correlation length
        icorr = self._nr - 1
        while np.absolute(self._kolmo_y[icorr]) < self._corrlim:
            icorr -= 1
        self._rcorr = self._kolmo_x[icorr]

        timer.stop()
        if self._rank == 0:
            log.debug("Initialized Kolmogorov in {} s".format(timer.seconds()))
        return

    def __repr__(self):
        ret = "<AtmSim rank = {}".format(self._rank)
        if self._rank == 0:
            ret += ":\n"
            ret += "  {} processes\n".format(self._ntask)
            ret += "  {} threads per process\n".format(self._nthread)
            ret += "  cachedir = {}\n".format(self._cachedir)
            ret += "  key1 = {}\n".format(self._key1)
            ret += "  key2 = {}\n".format(self._key2)
            ret += "  counter1 = {}\n".format(self._counter1)
            ret += "  counter2 = {}\n".format(self._counter2)
            ret += "  counter1start = {}\n".format(self._counter1start)
            ret += "  counter2start = {}\n".format(self._counter2start)
            ret += "  azmin = {}\n".format(self._azmin)
            ret += "  azmax = {}\n".format(self._azmax)
            ret += "  elmin = {}\n".format(self._elmin)
            ret += "  elmax = {}\n".format(self._elmax)
            ret += "  tmin = {}\n".format(self._tmin)
            ret += "  tmax = {}\n".format(self._tmax)
            ret += "  lmin_center = {}\n".format(self._lmin_center)
            ret += "  lmax_center = {}\n".format(self._lmax_center)
            ret += "  w_center = {}\n".format(self._w_center)
            ret += "  w_sigma = {}\n".format(self._w_sigma)
            ret += "  wdir_center = {}\n".format(self._wdir_center)
            ret += "  wdir_sigma = {}\n".format(self._wdir_sigma)
            ret += "  z0_center = {}\n".format(self._z0_center)
            ret += "  z0_sigma = {}\n".format(self._z0_sigma)
            ret += "  T0_center = {}\n".format(self._T0_center)
            ret += "  T0_sigma = {}\n".format(self._T0_sigma)
        ret += ">"
        return ret



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os
import shutil

import numpy as np
import numpy.testing as nt

import healpy as hp

from ..tod import AnalyticNoise, OpSimNoise, regular_intervals, OpFlagGaps
from ..todmap import TODSatellite, OpPointingHpix, OpMadam, OpAccumDiag

from ..map import DistPixels, covariance_invert, covariance_apply

from ._helpers import (
    create_outdir,
    create_distdata,
    boresight_focalplane,
    uniform_chunks,
)


class BinnedTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # a fixed number of detectors and one chunk per process.

        self.data = create_distdata(self.comm, obs_per_group=1)

        self.ndet = 1
        self.rate = 20.0
        self.hwprpm = 100

        # Create detectors with default properties
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet, samplerate=self.rate)

        # Samples per observation
        self.totsamp = 2000000

        # Pixelization

        self.sim_nside = 32
        self.sim_npix = 12 * self.sim_nside ** 2

        self.map_nside = 32
        self.map_npix = 12 * self.map_nside ** 2

        # Scan strategy

        self.spinperiod = 1.0
        self.spinangle = 30.0
        self.precperiod = 5.0
        self.precangle = 65.0

        # In order to make sure that the noise realization is reproducible
        # at all concurrencies, we set the chunksize to something independent
        # of the number of ranks.

        chunks = uniform_chunks(self.totsamp, nchunk=100)

        # define some valid data intervals so that we can test flag handling
        # in the gaps

        nint = 4
        intsamp = self.totsamp // nint
        inttime = (intsamp - 1) / self.rate
        durtime = inttime * 0.85
        gaptime = inttime - durtime
        intrvls = regular_intervals(nint, 0, 0, self.rate, durtime, gaptime)
        self.validsamp = 0
        for it in intrvls:
            self.validsamp += it.last - it.first + 1

        # We have multiple observations
        self.validsamp *= self.data.comm.ngroups

        # Populate the single observation per group

        tod = TODSatellite(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=1,
            firsttime=0.0,
            rate=self.rate,
            spinperiod=self.spinperiod,
            spinangle=self.spinangle,
            precperiod=self.precperiod,
            precangle=self.precangle,
            sampsizes=chunks,
            hwprpm=self.hwprpm,
        )

        tod.set_prec_axis()

        nse = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["noise"] = nse
        self.data.obs[0]["intervals"] = intrvls

    def test_binned(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # flag data outside valid intervals
        gapflagger = OpFlagGaps()
        gapflagger.exec(self.data)

        # generate noise timestreams from the noise model
        nsig = OpSimNoise()
        nsig.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True, mode="IQU")
        pointing.exec(self.data)

        handle = None
        if rank == 0:
            handle = open(os.path.join(self.outdir, "info.txt"), "w")
        self.data.info(handle=handle)
        if rank == 0:
            handle.close()

        # construct distributed maps to store the covariance,
        # noise weighted map, and hits

        invnpp = DistPixels(
            self.data, comm=self.data.comm.comm_world, nnz=6, dtype=np.float64
        )
        invnpp.data.fill(0.0)

        zmap = DistPixels(
            self.data, comm=self.data.comm.comm_world, nnz=3, dtype=np.float64
        )
        zmap.data.fill(0.0)

        hits = DistPixels(
            self.data, comm=self.data.comm.comm_world, nnz=1, dtype=np.int64
        )
        hits.data.fill(0)

        # accumulate the inverse covariance and noise weighted map.
        # Use detector weights based on the analytic NET.

        tod = self.data.obs[0]["tod"]
        nse = self.data.obs[0]["noise"]
        detweights = {}
        for d in tod.local_dets:
            detweights[d] = 1.0 / (self.rate * nse.NET(d) ** 2)

        build_invnpp = OpAccumDiag(
            detweights=detweights, invnpp=invnpp, hits=hits, zmap=zmap, name="noise"
        )
        build_invnpp.exec(self.data)

        invnpp.allreduce()
        hits.allreduce()
        zmap.allreduce()

        hits.write_healpix_fits(os.path.join(self.outdir, "hits.fits"))
        invnpp.write_healpix_fits(os.path.join(self.outdir, "invnpp.fits"))
        zmap.write_healpix_fits(os.path.join(self.outdir, "zmap.fits"))

        # invert it
        covariance_invert(invnpp, 1.0e-3)

        invnpp.write_healpix_fits(os.path.join(self.outdir, "npp.fits"))

        # compute the binned map, N_pp x Z

        covariance_apply(invnpp, zmap)
        zmap.write_healpix_fits(os.path.join(self.outdir, "binned.fits"))

        # compare with MADAM

        madam_out = os.path.join(self.outdir, "madam")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["temperature_only"] = "F"
        pars["force_pol"] = "T"
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = self.map_nside
        pars["pixlim_cross"] = 1.0e-2
        pars["pixlim_map"] = 1.0e-3
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "T"
        pars["write_wcov"] = "T"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(params=pars, detweights=detweights, name="noise")

        if madam.available:
            madam.exec(self.data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True, verbose=False)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                toastfile = os.path.join(self.outdir, "hits.fits")
                toasthits = hp.read_map(toastfile, nest=True, verbose=False)

                nt.assert_equal(hits, toasthits)

                tothits = np.sum(hits)
                nt.assert_equal(self.validsamp, tothits)

                covfile = os.path.join(madam_out, "madam_wcov_inv.fits")
                cov = hp.read_map(covfile, nest=True, field=None, verbose=False)

                toastfile = os.path.join(self.outdir, "invnpp.fits")
                toastcov = hp.read_map(toastfile, nest=True, field=None, verbose=False)

                """
                for p in range(6):
                    print("elem {} madam min/max = ".format(p),
                          np.min(cov[p]), " / ", np.max(cov[p]))
                    print("elem {} toast min/max = ".format(p),
                          np.min(toastcov[p]), " / ", np.max(toastcov[p]))
                    print("elem {} invNpp max diff = ".format(p),
                          np.max(np.absolute(toastcov[p] - cov[p])))
                    nt.assert_almost_equal(cov[p], toastcov[p])
                """

                covfile = os.path.join(madam_out, "madam_wcov.fits")
                cov = hp.read_map(covfile, nest=True, field=None, verbose=False)

                toastfile = os.path.join(self.outdir, "npp.fits")
                toastcov = hp.read_map(toastfile, nest=True, field=None, verbose=False)

                """
                for p in range(6):
                    covdiff = toastcov[p] - cov[p]
                    print("elem {} madam min/max = ".format(p),
                          np.min(cov[p]), " / ", np.max(cov[p]))
                    print("elem {} toast min/max = ".format(p),
                          np.min(toastcov[p]), " / ", np.max(toastcov[p]))
                    print("elem {} Npp max diff = ".format(p),
                          np.max(np.absolute(covdiff[p])))
                    print("elem {} Npp mean / rms diff = ".format(p),
                          np.mean(covdiff[p]), " / ", np.std(covdiff[p]))
                    print("elem {} Npp relative diff mean / rms = ".format(p),
                          np.mean(np.absolute(covdiff[p]/cov[p])), " / ",
                          np.std(np.absolute(covdiff[p]/cov[p])))
                    nt.assert_almost_equal(cov[p], toastcov[p])
                """

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True, field=None, verbose=False)
                mask = hp.mask_bad(bins[0])
                bins[0][mask] = 0.0
                mask = hp.mask_bad(bins[1])
                bins[1][mask] = 0.0
                mask = hp.mask_bad(bins[2])
                bins[2][mask] = 0.0

                outfile = "{}_I.png".format(binfile)
                hp.mollview(bins[0], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                outfile = "{}_Q.png".format(binfile)
                hp.mollview(bins[1], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                outfile = "{}_U.png".format(binfile)
                hp.mollview(bins[2], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                toastfile = os.path.join(self.outdir, "binned.fits")
                toastbins = hp.read_map(toastfile, nest=True, field=None, verbose=False)

                outfile = "{}_I.png".format(toastfile)
                hp.mollview(toastbins[0], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                outfile = "{}_Q.png".format(toastfile)
                hp.mollview(toastbins[1], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                outfile = "{}_U.png".format(toastfile)
                hp.mollview(toastbins[2], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to madam output

                diffmap = toastbins[0] - bins[0]
                mask = bins[0] != 0
                """
                print("toast/madam I diff mean / std = ",
                      np.mean(diffmap[mask]), np.std(diffmap[mask]))
                print("toast/madam I diff rel ratio min / max = ",
                      np.min(diffmap[mask]/bins[0][mask]), " / ",
                      np.max(diffmap[mask]/bins[0][mask]))
                """
                outfile = "{}_diff_madam_I.png".format(toastfile)
                hp.mollview(diffmap, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()
                # nt.assert_almost_equal(bins[0][mask], binserial[0][mask],
                #                       decimal=6)

                diffmap = toastbins[1] - bins[1]
                mask = bins[1] != 0
                """
                print("toast/madam Q diff mean / std = ",
                      np.mean(diffmap[mask]), np.std(diffmap[mask]))
                print("toast/madam Q diff rel ratio min / max = ",
                      np.min(diffmap[mask]/bins[1][mask]), " / ",
                      np.max(diffmap[mask]/bins[1][mask]))
                """
                outfile = "{}_diff_madam_Q.png".format(toastfile)
                hp.mollview(diffmap, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()
                # nt.assert_almost_equal(bins[1][mask], binserial[1][mask],
                #                       decimal=6)

                diffmap = toastbins[2] - bins[2]
                mask = bins[2] != 0
                """
                print("toast/madam U diff mean / std = ",
                      np.mean(diffmap[mask]), np.std(diffmap[mask]))
                print("toast/madam U diff rel ratio min / max = ",
                      np.min(diffmap[mask]/bins[2][mask]), " / ",
                      np.max(diffmap[mask]/bins[2][mask]))
                """
                outfile = "{}_diff_madam_U.png".format(toastfile)
                hp.mollview(diffmap, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()
                # nt.assert_almost_equal(bins[2][mask], binserial[2][mask],
                #                       decimal=6)

                # compute the binned map serially as a check

                zfile = os.path.join(self.outdir, "zmap.fits")
                ztoast = hp.read_map(zfile, nest=True, field=None, verbose=False)

                binserial = np.copy(ztoast)
                for p in range(self.map_npix):
                    binserial[0][p] = (
                        toastcov[0][p] * ztoast[0][p]
                        + toastcov[1][p] * ztoast[1][p]
                        + toastcov[2][p] * ztoast[2][p]
                    )
                    binserial[1][p] = (
                        toastcov[1][p] * ztoast[0][p]
                        + toastcov[3][p] * ztoast[1][p]
                        + toastcov[4][p] * ztoast[2][p]
                    )
                    binserial[2][p] = (
                        toastcov[2][p] * ztoast[0][p]
                        + toastcov[4][p] * ztoast[1][p]
                        + toastcov[5][p] * ztoast[2][p]
                    )

                toastfile = os.path.join(self.outdir, "binned_serial")
                outfile = "{}_I.png".format(toastfile)
                hp.mollview(binserial[0], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                outfile = "{}_Q.png".format(toastfile)
                hp.mollview(binserial[1], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                outfile = "{}_U.png".format(toastfile)
                hp.mollview(binserial[2], xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to madam output

                diffmap = binserial[0] - bins[0]
                mask = bins[0] != 0
                """
                print("serial/madam I diff mean / std = ",
                      np.mean(diffmap[mask]), np.std(diffmap[mask]))
                print("serial/madam I diff rel ratio min / max = ",
                      np.min(diffmap[mask]/bins[0][mask]), " / ",
                      np.max(diffmap[mask]/bins[0][mask]))
                """
                outfile = "{}_diff_madam_I.png".format(toastfile)
                hp.mollview(diffmap, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()
                nt.assert_almost_equal(bins[0][mask], binserial[0][mask], decimal=3)

                diffmap = binserial[1] - bins[1]
                mask = bins[1] != 0
                """
                print("serial/madam Q diff mean / std = ",
                      np.mean(diffmap[mask]), np.std(diffmap[mask]))
                print("serial/madam Q diff rel ratio min / max = ",
                      np.min(diffmap[mask]/bins[1][mask]), " / ",
                      np.max(diffmap[mask]/bins[1][mask]))
                """
                outfile = "{}_diff_madam_Q.png".format(toastfile)
                hp.mollview(diffmap, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()
                nt.assert_almost_equal(bins[1][mask], binserial[1][mask], decimal=3)

                diffmap = binserial[2] - bins[2]
                mask = bins[2] != 0
                """
                print("serial/madam U diff mean / std = ",
                      np.mean(diffmap[mask]), np.std(diffmap[mask]))
                print("serial/madam U diff rel ratio min / max = ",
                      np.min(diffmap[mask]/bins[2][mask]), " / ",
                      np.max(diffmap[mask]/bins[2][mask]))
                """
                outfile = "{}_diff_madam_U.png".format(toastfile)
                hp.mollview(diffmap, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()
                nt.assert_almost_equal(bins[2][mask], binserial[2][mask], decimal=3)

        else:
            if rank == 0:
                print("libmadam not available, skipping tests", flush=True)

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from .mpi import MPITestCase

import numpy.testing as nt

from ..tod.interval import intervals_to_chunklist

from ..tod.sim_interval import regular_intervals


class IntervalTest(MPITestCase):
    def setUp(self):
        self.rate = 123.456
        self.duration = 24 * 3601.23
        self.gap = 3600.0
        self.start = 5432.1
        self.first = 10
        self.nint = 3

    def test_tochunks(self):
        intrvls = regular_intervals(
            self.nint, self.start, self.first, self.rate, self.duration, self.gap
        )
        totsamp = self.nint * (intrvls[0].last - intrvls[0].first + 1)
        totsamp += self.nint * (intrvls[1].first - intrvls[0].last - 1)
        sizes = intervals_to_chunklist(intrvls, totsamp, startsamp=self.first + 10)
        # for it in intrvls:
        #     print(it.first," ",it.last," ",it.start," ",it.stop)
        # print(sizes)
        nt.assert_equal(np.sum(sizes), totsamp)

    def test_regular(self):
        intrvls = regular_intervals(
            self.nint, self.start, self.first, self.rate, self.duration, self.gap
        )

        goodsamp = self.nint * (int(self.duration * self.rate) + 1)

        check = 0

        for it in intrvls:
            # print(it.first," ",it.last," ",it.start," ",it.stop)
            check += it.last - it.first + 1

        nt.assert_equal(check, goodsamp)
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import sys

import numpy as np

import ctypes

from .mpi import MPITestCase

from ..cache import Cache

from ..utils import AlignedF64, memreport


class CacheTest(MPITestCase):
    def setUp(self):
        self.nsamp = 1000
        self.membytes = 100000000
        self.memnbuf = 100
        self.cache = Cache(pymem=False)
        self.pycache = Cache(pymem=True)
        self.types = {
            "f64": np.float64,
            "f32": np.float32,
            "i64": np.int64,
            "u64": np.uint64,
            "i32": np.int32,
            "u32": np.uint32,
            "i16": np.int16,
            "u16": np.uint16,
            "i8": np.int8,
            "u8": np.uint8,
        }

    def tearDown(self):
        del self.cache
        del self.pycache

    def test_dict_interface(self):
        cache = Cache()
        cache["aaa"] = np.arange(10)
        cache["ccc"] = np.arange(10) + 10
        self.assertTrue("aaa" in cache)
        self.assertFalse("bbb" in cache)
        self.assertEqual(len(cache), 2)
        n = 0
        for x in cache:
            n += 1
        self.assertEqual(len(cache), n)
        del cache["aaa"]
        self.assertEqual(len(cache), n - 1)
        return

    def test_refcount(self):
        buf = AlignedF64.zeros(self.nsamp)
        # print("REFCNT: ", sys.getrefcount(buf), flush=True)
        # wbuf1 = buf.weak()
        # print("REFCNT 1: ", sys.getrefcount(buf), flush=True)
        # wbuf2 = buf.weak()
        # print("REFCNT 2: ", sys.getrefcount(buf), flush=True)

        return

    def test_create(self):
        for k, v in self.types.items():
            ref = self.cache.create("test-{}".format(k), v, (self.nsamp, 4))
            del ref

        for k, v in self.types.items():
            data = self.cache.reference("test-{}".format(k))
            data[:] += np.repeat(np.arange(self.nsamp, dtype=v), 4).reshape(-1, 4)
            del data

        for k, v in self.types.items():
            ex = self.cache.exists("test-{}".format(k))
            self.assertTrue(ex)

        for k, v in self.types.items():
            self.cache.destroy("test-{}".format(k))

        self.cache.clear()

        for k, v in self.types.items():
            ref = self.pycache.create("test-{}".format(k), v, (self.nsamp, 4))
            del ref

        for k, v in self.types.items():
            data = self.pycache.reference("test-{}".format(k))
            data[:] += np.repeat(np.arange(self.nsamp, dtype=v), 4).reshape(-1, 4)
            del data

        for k, v in self.types.items():
            ex = self.pycache.exists("test-{}".format(k))
            self.assertTrue(ex)

        for k, v in self.types.items():
            self.pycache.destroy("test-{}".format(k))

        self.pycache.clear()
        return

    def test_put(self):
        # Populate some buffers
        for k, v in self.types.items():
            pname = "test-{}".format(k)
            pdata = np.ones((self.nsamp, 4), v)
            ref = self.cache.put(pname, pdata)
            del ref

        # Test putting the same buffer- it should just return a new array
        # wrapping the same underlying memory.
        for k, v in self.types.items():
            pname = "test-{}".format(k)
            pdata = np.ones((self.nsamp, 4), v)
            ref = self.cache.reference(pname)
            copyref = self.cache.put(pname, ref)
            ref_pnt = ref.ctypes.data_as(ctypes.c_void_p).value
            copyref_pnt = copyref.ctypes.data_as(ctypes.c_void_p).value
            self.assertTrue(ref_pnt == copyref_pnt)
            del ref
            del copyref

        # Test replacement of buffers with the same name
        for k, v in self.types.items():
            pname = "test-{}".format(k)
            pdata = np.ones((self.nsamp, 8), v)
            try:
                # This should raise- same name as existing buffer, but
                # replace == False
                ref = self.cache.put(pname, pdata)
                del ref
                self.assertTrue(False)
            except RuntimeError:
                # Success!
                pass
            # This should work
            ref = self.cache.put(pname, pdata, replace=True)
            del ref

        self.cache.clear()

        # Populate some buffers
        for k, v in self.types.items():
            pname = "test-{}".format(k)
            pdata = np.ones((self.nsamp, 4), v)
            ref = self.pycache.put(pname, pdata)
            del ref

        # Test putting the same buffer- it should just return a new array
        # wrapping the same underlying memory.
        for k, v in self.types.items():
            pname = "test-{}".format(k)
            pdata = np.ones((self.nsamp, 4), v)
            ref = self.pycache.reference(pname)
            copyref = self.pycache.put(pname, ref)
            ref_pnt = ref.ctypes.data_as(ctypes.c_void_p).value
            copyref_pnt = copyref.ctypes.data_as(ctypes.c_void_p).value
            self.assertTrue(ref_pnt == copyref_pnt)
            del ref
            del copyref

        # Test replacement of buffers with the same name
        for k, v in self.types.items():
            pname = "test-{}".format(k)
            pdata = np.ones((self.nsamp, 8), v)
            try:
                # This should raise- same name as existing buffer, but
                # replace == False
                ref = self.pycache.put(pname, pdata)
                del ref
                self.assertTrue(False)
            except RuntimeError:
                # Success!
                pass
            # This should work
            ref = self.pycache.put(pname, pdata, replace=True)
            del ref

        self.pycache.clear()
        return

    def test_create_none(self):
        try:
            ref = self.cache.create(None, np.float, (1, 10))
            del ref
            ref = self.pycache.create(None, np.float, (1, 10))
            del ref
            raise RuntimeError("Creating object with None key succeeded")
        except ValueError:
            pass

        self.cache.clear()
        self.pycache.clear()
        return

    def test_put_none(self):
        try:
            ref = self.cache.put(None, np.float, np.arange(10))
            del ref
            ref = self.pycache.put(None, np.float, np.arange(10))
            del ref
            raise RuntimeError("Putting an object with None key succeeded")
        except ValueError:
            pass

        self.cache.clear()
        self.pycache.clear()
        return

    def test_clear(self):
        for k, v in self.types.items():
            ref = self.cache.create("test-{}".format(k), v, (self.nsamp, 4))
            del ref
            ref = self.pycache.create("test-{}".format(k), v, (self.nsamp, 4))
            del ref
        self.cache.clear()
        self.pycache.clear()
        return

    def test_alias(self):
        ref = self.cache.put("test", np.arange(10))
        del ref

        self.cache.add_alias("test-alias", "test")
        self.cache.add_alias("test-alias-2", "test")

        data = self.cache.reference("test-alias")
        del data

        self.cache.destroy("test-alias")

        data = self.cache.reference("test-alias-2")
        del data

        self.cache.destroy("test")

        ex = self.cache.exists("test-alias-2")
        self.assertFalse(ex)

        self.cache.clear()

        ref = self.pycache.put("test", np.arange(10))
        del ref

        self.pycache.add_alias("test-alias", "test")
        self.pycache.add_alias("test-alias-2", "test")

        data = self.pycache.reference("test-alias")
        del data

        self.pycache.destroy("test-alias")

        data = self.pycache.reference("test-alias-2")
        del data

        self.pycache.destroy("test")

        ex = self.pycache.exists("test-alias-2")
        self.assertFalse(ex)

        self.pycache.clear()
        return

    def test_memfree(self):
        memreport(comm=self.comm, msg="Before large buffer creation")
        memcache = Cache(pymem=False)
        ref = memcache.create("test-big", np.uint8, (self.membytes,))
        del ref

        memreport(comm=self.comm, msg="After large buffer creation")

        mem = memcache.report(silent=True)
        self.assertEqual(mem, self.membytes)
        print("Cache now has {} bytes".format(mem), flush=True)

        memcache.clear()

        memreport(comm=self.comm, msg="After cache clear")
        mem = memcache.report(silent=True)
        self.assertEqual(mem, 0)
        print("Cache now has {} bytes".format(mem), flush=True)

        smallbytes = self.membytes // self.memnbuf
        for i in range(self.memnbuf):
            name = "test-small_{}".format(i)
            ref = memcache.create(name, np.uint8, (smallbytes,))
            del ref

        memreport(
            comm=self.comm,
            msg="After creation of {} small buffers".format(self.memnbuf),
        )
        mem = memcache.report(silent=True)
        self.assertEqual(mem, self.memnbuf * smallbytes)
        print("Cache now has {} bytes".format(mem), flush=True)

        memcache.clear()
        memreport(comm=self.comm, msg="After cache clear")
        mem = memcache.report(silent=True)
        self.assertEqual(mem, 0)
        print("Cache now has {} bytes".format(mem), flush=True)
